import gfapy
import re

def unsafe_decode(string):
  return gfapy.LastPos(string)

def decode(string):
  position = unsafe_decode(string)
  value = gfapy.posvalue(position)
  if value < 0:
    raise gfapy.ValueError(
      "{} is not a positive integer".format(value))
  return position

def validate_decoded(obj):
  if isinstance(obj, int):
    if obj < 0:
      raise gfapy.ValueError(
        "{} is not a positive integer".format(obj))
  elif isinstance(obj, gfapy.LastPos):
    obj.validate()
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: int, gfapy.LastPos)")

def validate_encoded(string):
  if not re.match(r"^[0-9]+\$?$", string):
    raise gfapy.FormatError(
      "{} is not a valid GFA2 position\n".format(repr(string))+
      "(it must be an unsigned integer eventually followed by a $)")

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
  else:
    validate_decoded(obj)
  return str(obj)



import gfapy

def invert(symbol):
  """Computes the inverted orientation or end_type symbol.

  Parameters:
    symbol (str) : a one-character string, symbolizing an orientation (+ or -)
      or an end-type (L or R)

  Returns:
    str : the other one character string of the same category (e.g. - for +)

  Raises:
    gfapy.error.ValueError : if a string other than the mentioned ones is used
  """
  if symbol == "+":
    return "-"
  elif symbol == "-":
    return "+"
  elif symbol == "L":
    return "R"
  elif symbol == "R":
    return "L"
  else:
    raise gfapy.ValueError("No inverse defined for {}".format(symbol))



import gfapy

class Finders:
  def segment(self, s):
    """Search a segment in a GFA.

    If the argument is a line, it is returned. If it is a string,
    it is used as a segment identifier, and the segment with that identifier
    is returned. If no segment has the identifier, None is returned.
    Note that None is also returned if a line of another type (not segment)
    has the identifier (differently from the line() method, which returns
    the lines independently from the record type).

    Parameters:
      l (str, gfapy.Line)
    """
    if isinstance(s, gfapy.Line):
      return s
    else:
      return self._records["S"].get(s, None)

  def try_get_segment(self, s):
    """Call segment() and raise an exception is the segment is not found."""
    seg = self.segment(s)
    if seg is None:
      raise gfapy.NotFoundError("No segment has name {}".format(s))
    else:
      return seg

  RECORDS_WITH_NAME = ["E", "S", "P", "U", "G", "O", "\n"]

  def line(self, l):
    """Search a line in a GFA.

    If the argument is a line, it is returned. If it is a string,
    it is used as a line identifier, and the line with that identifier
    is returned. If no line has the identifier, None is returned.

    Parameters:
      l (str, gfapy.Line)
    """
    if gfapy.is_placeholder(l):
      return None
    elif isinstance(l, gfapy.Line):
      return l
    elif isinstance(l, str):
      return self.__line_by_name(l)
    else:
      return None

  def try_get_line(self, l):
    """Call line() and raise an exception is the line is not found."""
    gfa_line = self.line(l)
    if gfa_line is None:
      if gfapy.is_placeholder(l):
        raise gfapy.ValueError(
          "'*' is a placeholder and not a valid name for a line")
      else:
        raise gfapy.NotFoundError(
            "No line found with ID {}".format(l))
    return gfa_line

  def fragments_for_external(self, external_id):
    """List of F lines with a given external ID."""
    return list(self._records["F"].get(external_id,{}).values())

  def select(self, dict_or_line):
    """Select all lines which respect a chriterion.

    The chriterion is expressed by the argument, which is either a line
    instance or a dictionary. If it is a dictionary, it shall contain
    pairs of fieldnames/values and the method returns all lines
    where the mentioned fieldnames have the corresponding values.
    If it is a line, it is compared with the lines of the same type
    in the Gfa instance and lines with the same field values
    are returned (undefined placeholder values are thereby not compared).
    """
    is_dict = isinstance(dict_or_line, dict)
    name = dict_or_line.get("name",None) if is_dict else dict_or_line.get("name")
    if name is not None and not gfapy.is_placeholder(name):
      collection = [self.__line_by_name(name)]
    else:
      if is_dict:
        record_type = dict_or_line.get("record_type",None)
      else:
        record_type = dict_or_line.record_type
      collection = self.__collection_for_select(record_type)
    method = "_has_field_values" if is_dict else "_has_eql_fields"
    return [line for line in collection \
        if getattr(line, method)(dict_or_line, ["record_type","name"])]

  def _search_duplicate(self, gfa_line):
    if gfa_line.record_type == "L":
      return self._search_link(gfa_line.oriented_from, gfa_line.oriented_to,
                               gfa_line.alignment)
    elif gfa_line.record_type in self.RECORDS_WITH_NAME:
      return self.line(gfa_line.name)
    else:
      return None

  def _search_link(self, orseg1, orseg2, cigar):
    s = self.segment(orseg1.line)
    if s is None:
      return None
    for l in s.dovetails:
      if isinstance(l, gfapy.line.edge.Link) and \
          l.is_compatible(orseg1, orseg2, cigar, True):
        return l
    return None

  def __line_by_name(self, name):
    for rt in self.RECORDS_WITH_NAME:
      if rt not in self._records:
        next
      found = self._records[rt].get(name, None)
      if found is not None:
        return found
    return None

  def __collection_for_select(self, record_type):
    if record_type is None:
      return self.lines
    else:
      d = self._records[record_type]
      if record_type == "F":
        retval = []
        for v in d.values():
          retval.extend(list(v.values()))
        return retval
      else:
        return list(d.values())



import gfapy

class References:

  def add_item(self, item):
    """
    Add an item to the group.

    Parameters
    ----------
    item : gfapy.Line, str
      GFA2 edge, segment, gap or group line to add.
    """
    if not self.is_connected():
      self._add_item_to_unconnected_group(item)
    else:
      self._add_item_to_connected_group(item)

  def rm_item(self, item):
    """
    Remove an item from the group.

    Parameters
    ----------
    item : str, gfapy.Line
      GFA2 edge, segment, gap or group line to remove.
    """
    if not self.is_connected():
      self._rm_item_from_unconnected_group(item)
    else:
      self._rm_item_from_connected_group(item)

  def _rm_item_from_unconnected_group(self, item):
    if isinstance(item, gfapy.Line):
      item = item.name
    self._check_item_included(item)
    self.items.delete(item)
    return None

  def _rm_item_from_connected_group(self, item):
    if isinstance(item, str):
      item = self._gfa.line(item)
    self._check_item_included(item)
    item._delete_reference(self, "sets")
    self._delete_reference(item, "items")
    return None

  def _check_item_included(self, item):
    if item not in self.items:
      raise gfapy.NotFoundError(
        "Line: {}\n".format(self)+
        "Item: {}".format(repr(item))+
        "Items of the line do not include the item")

  def _add_item_to_unconnected_group(self, item, append = True):
    if isinstance(item, gfapy.Line):
      item = item.name
    if append:
      self.items.append(item)
    else:
      self.items.insert(0, item)
    return None

  def _add_item_to_connected_group(self, item, append = True):
    self._add_reference(self.prepare_and_check_ref(item),
                       "items", append = append)
    return None

  def _initialize_references(self):
    for i in range(len(self.items)):
      self.items[i] = self._line_for_ref_symbol(self.items[i])



from ..gfa2.references import References as  GFA2_References
from ..gfa2.same_id import SameID
from .references import References as Ordered_References
from .captured_path import CapturedPath
from .to_gfa1 import ToGFA1
from .. import Group

class Ordered(Ordered_References, CapturedPath, GFA2_References, SameID,
              ToGFA1, Group):
  """
  An ordered group line of a GFA2 file
  """

  RECORD_TYPE = "O"
  POSFIELDS = ["pid", "items"]
  FIELD_ALIAS = { "name" : "pid" }
  DATATYPE = {
    "pid" : "optional_identifier_gfa2",
    "items" : "oriented_identifier_list_gfa2",
  }
  NAME_FIELD = "pid"
  REFERENCE_FIELDS = ["items"]
  DEPENDENT_LINES = ["paths", "sets"]

Ordered._apply_definitions()



import gfapy

class OrientedSegments:

  @property
  def oriented_from(self):
    """
    Returns
    -------
    gfapy.OrientedLine
      The oriented segment represented by the from_segment/from_orient fields.
    """
    return gfapy.OrientedLine(self.from_segment, self.from_orient)

  @property
  def oriented_to(self):
    """
    Returns
    -------
    gfapy.OrientedLine
      The oriented segment represented by the to_segment/to_orient fields.
    """
    return gfapy.OrientedLine(self.to_segment, self.to_orient)



import gfapy

def unsafe_decode(string):
  return string

def decode(string):
  validate_decoded(string)
  return string

def validate_decoded(string):
  if string != "+" and string != "-":
    raise gfapy.FormatError(
      "{} is not a valid orientation\n".format(repr(string))+
      "(it must be + or -)")
  return string

#identical to validate_decoded, because python version uses strings for symbols
def validate_encoded(string):
  if string != "+" and string != "-":
    raise gfapy.FormatError(
      "{} is not a valid orientation\n".format(repr(string))+
      "(it must be + or -)")
  return string

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str)")



import gfapy
import unittest

class TestAPILinesCollections(unittest.TestCase):

  def test_gfa1_collections(self):
    gfa = gfapy.Gfa.from_file("tests/testdata/all_line_types.gfa1.gfa")
    # comments
    self.assertEqual(1, len(gfa.comments))
    self.assertRegex(gfa.comments[0].content, r'collections')
    # containments
    self.assertEqual(2, len(gfa.containments))
    self.assertEqual({"2_to_6", "1_to_5"}, {x.name for x in gfa.containments})
    # dovetails
    self.assertEqual(4, len(gfa.dovetails))
    self.assertEqual(set(["1_to_2", "1_to_3", "11_to_12", "11_to_13"]),
                 set([x.name for x in gfa.dovetails]))
    # edges
    self.assertEqual(6, len(gfa.edges))
    self.assertEqual(set(["1_to_2", "1_to_3", "11_to_12",
                  "11_to_13", "2_to_6", "1_to_5"]),
                 set([x.name for x in gfa.edges]))
    # segments
    self.assertSetEqual(set(["1", "3", "5", "13", "11", "12", "4", "6", "2"]),
                 set([x.name for x in gfa.segments]))
    # segment_names
    self.assertSetEqual(set(["1", "3", "5", "13", "11", "12", "4", "6", "2"]),
                 set(gfa.segment_names))
    # paths
    self.assertSetEqual(set(["14", "15"]), set([x.name for x in gfa.paths]))
    # path_names
    self.assertSetEqual(set(["14", "15"]), set(gfa.path_names))
    # names
    self.assertSetEqual(set(gfa.segment_names + gfa.path_names +
                        gfa.edge_names), set(gfa.names))
    # lines
    self.assertEqual(set([str(x) for x in gfa.comments + gfa.headers + gfa.segments + gfa.edges +
                 gfa.paths]), set([str(x) for x in gfa.lines]))

  def test_gfa2_collections(self):
    gfa = gfapy.Gfa.from_file("tests/testdata/all_line_types.gfa2.gfa")
    # comments
    self.assertEqual(3, len(gfa.comments))
    self.assertRegex(gfa.comments[0].content, r'collections')
    # edges
    self.assertSetEqual(set(["1_to_2", "2_to_6", "1_to_3",
                  "11_to_12", "11_to_13", "1_to_5"]),
                 set([x.name for x in gfa.edges]))
    # edge_names
    self.assertSetEqual(set(["1_to_2", "2_to_6", "1_to_3",
                  "11_to_12", "11_to_13", "1_to_5"]),
                 set(gfa.edge_names))
    # dovetails
    self.assertSetEqual(set(["1_to_2", "1_to_3", "11_to_12", "11_to_13"]),
                 set([x.name for x in gfa.dovetails]))
    # containments
    self.assertSetEqual(set(["2_to_6", "1_to_5"]),
                 set([x.name for x in gfa.containments]))
    # gaps
    self.assertSetEqual(set(["1_to_11", "2_to_12"]), set([x.name for x in gfa.gaps]))
    # gap_names
    self.assertSetEqual(set(["1_to_11", "2_to_12"]), set(gfa.gap_names))
    # sets
    self.assertSetEqual(set(["16", "16sub"]), set([x.name for x in gfa.sets]))
    # set_names
    self.assertSetEqual(set(["16", "16sub"]), set(gfa.set_names))
    # paths
    self.assertSetEqual(set(["14", "15"]), set([x.name for x in gfa.paths]))
    # path_names
    self.assertSetEqual(set(["14", "15"]), set(gfa.path_names))
    # segments
    self.assertSetEqual(set(["1", "3", "5", "13", "11", "12", "4", "6", "2"]),
               set([x.name for x in gfa.segments]))
    # segment_names
    self.assertSetEqual(set(["1", "3", "5", "13", "11", "12", "4", "6", "2"]),
               set(gfa.segment_names))
    # fragments
    self.assertSetEqual(set(["read1_in_2", "read2_in_2"]),
        set([x.get("id") for x in gfa.fragments]))
    # external_names
    self.assertSetEqual(set(["read1", "read2"]), set(gfa.external_names))
    # custom_record_keys
    self.assertSetEqual(set(["X", "Y"]), set(gfa.custom_record_keys))
    # custom_records
    self.assertEqual(3, len(gfa.custom_records))
    self.assertSetEqual(set(["X", "X", "Y"]), set([x.record_type for x in gfa.custom_records]))
    # custom_records("X")
    self.assertSetEqual(set(["X", "X"]), set([x.record_type for x in gfa.custom_records_of_type("X")]))
    # names
    self.assertSetEqual(set(gfa.segment_names + gfa.edge_names + gfa.gap_names +
                 gfa.path_names + gfa.set_names), set(gfa.names))
    # lines
    self.assertSetEqual(set([str(x) for x in gfa.comments + gfa.headers + gfa.segments + gfa.edges +
                 gfa.paths + gfa.sets + gfa.gaps + gfa.fragments +
                 gfa.custom_records]), set([str(x) for x in gfa.lines]))



import gfapy
import re

def unsafe_decode(string):
  return [ gfapy.Alignment(s, version = "gfa1", valid = True) \
             for s in string.split(",") ]

def decode(string):
  validate_encoded(string)
  return unsafe_decode(string)

def validate_encoded(string):
  if not re.match(r"^(\*|(([0-9]+[MIDNSHPX=])+))(,(\*|(([0-9]+[MIDNSHPX=])+)))*$", string):
     raise gfapy.FormatError(
       "{} is not a comma separated list of * or CIGARs\n".format(repr(string))+
       "(CIGAR strings must match ([0-9]+[MIDNSHPX=])+)")

def validate_decoded(obj):
  if isinstance(obj, gfapy.Placeholder):
    pass
  elif isinstance(obj, list):
    for e in obj:
      e = gfapy.Alignment(e, version = "gfa1")
      e.validate()
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n".format(obj.__class__.__name__)+
      "(accepted classes: list, AlignmentPlaceholder)")

def unsafe_encode(obj):
  if isinstance(obj, gfapy.Placeholder):
    return str(obj)
  elif isinstance(obj, list):
    return ",".join([str(gfapy.Alignment(cig)) for cig in obj])
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: list, AlignmentPlaceholder)")

def encode(obj):
  if isinstance(obj, gfapy.Placeholder):
    return str(obj)
  if isinstance(obj, list):
    def f(cig):
      cig = gfapy.Alignment(cig)
      cig.validate()
      return str(cig)
    return ",".join(map(f, obj))
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: list, AlignmentPlaceholder)")



import unittest
from gfapy import *

class TestApiPlaceholders(unittest.TestCase):

  def test_str(self):
    self.assertEqual("*", str(Placeholder()))

  def test_is_placeholder(self):
    self.assertTrue(is_placeholder(Placeholder()))
    self.assertTrue(is_placeholder("*"))
    self.assertTrue(is_placeholder([]))
    self.assertFalse(is_placeholder("a"))
    self.assertFalse(is_placeholder("**"))
    self.assertFalse(is_placeholder(1))
    self.assertFalse(is_placeholder(1.0))
    self.assertFalse(is_placeholder(["x"]))

  def test_compatibility_methods(self):
    p = Placeholder()
    self.assertTrue(p.is_empty())
    self.assertTrue(is_placeholder(p))
    self.assertEqual(0, len(p))
    self.assertTrue(is_placeholder(p.rc()))
    self.assertTrue(is_placeholder(p + 1))
    self.assertTrue(is_placeholder(p[0]))



import unittest
import gfapy

class TestUnitLineFinders(unittest.TestCase):


  l_gfa1_str = ["S\t1\t*",
                "S\t2\t*",
                "S\t3\t*",
                "S\t4\tCGAT",
                "L\t1\t+\t2\t+\t*",
                "L\t1\t-\t3\t+\t10M",
                "C\t1\t-\t4\t-\t1\t*",
                "P\tp1\t1+,2+\t*"]
  l_gfa1 = [gfapy.Line(s) for s in l_gfa1_str]
  gfa1 = gfapy.Gfa(l_gfa1)

  l_gfa2_str = ["S\t5\t100\t*",
                "S\t6\t110\t*",
                "E\te1\t5+\t6-\t0\t100$\t10\t110$\t*",
                "G\tg1\t5-\t6-\t1000\t*",
                "O\to1\t5+ 6-",
                "U\tu1\t5 e1",
                "F\t5\tread1-\t0\t10\t102\t122\t*",
                "F\t5\tread1-\t30\t100$\t180\t255\t*",
                "F\t6\tread1-\t40\t50\t52\t64\t*",
                "X\tx1\txx:Z:A",
                "X\tx2",
                "G\t*\t5+\t6+\t2000\t*"]
  l_gfa2 = [gfapy.Line(s) for s in l_gfa2_str]
  gfa2 = gfapy.Gfa(l_gfa2)

  def test_search_link(self):
    # search using the direct link
    self.assertEqual(TestUnitLineFinders.l_gfa1[4], TestUnitLineFinders.gfa1._search_link(gfapy.OrientedLine("1","+"), gfapy.OrientedLine("2","+"), "*"))
    # search using the complement link
    self.assertEqual(TestUnitLineFinders.l_gfa1[4], TestUnitLineFinders.gfa1._search_link(gfapy.OrientedLine("2","-"), gfapy.OrientedLine("1","-"), "*"))
    # with cigar parameter, but placeholder in line
    self.assertEqual(TestUnitLineFinders.l_gfa1[4],
                 TestUnitLineFinders.gfa1._search_link(gfapy.OrientedLine("1","+"), gfapy.OrientedLine("2","+"), "10M"))
    # with cigar parameter, and cigar in line
    self.assertEqual(TestUnitLineFinders.l_gfa1[5],
                 TestUnitLineFinders.gfa1._search_link(gfapy.OrientedLine("1","-"), gfapy.OrientedLine("3","+"), "10M"))
    self.assertEqual(None,
                 TestUnitLineFinders.gfa1._search_link(gfapy.OrientedLine("1","-"), gfapy.OrientedLine("3","+"), "12M"))
    # with placeholder parameter, and cigar in line
    self.assertEqual(TestUnitLineFinders.l_gfa1[5],
                 TestUnitLineFinders.gfa1._search_link(gfapy.OrientedLine("1","-"), gfapy.OrientedLine("3","+"), "*"))

  def test_search_duplicate_gfa1(self):
    # link
    self.assertEqual(TestUnitLineFinders.l_gfa1[4], TestUnitLineFinders.gfa1._search_duplicate(TestUnitLineFinders.l_gfa1[4]))
    # complement link
    self.assertEqual(TestUnitLineFinders.l_gfa1[4], TestUnitLineFinders.gfa1._search_duplicate(TestUnitLineFinders.l_gfa1[4].complement()))
    # containment
    self.assertEqual(None, TestUnitLineFinders.gfa1._search_duplicate(TestUnitLineFinders.l_gfa1[6]))
    # segment
    self.assertEqual(TestUnitLineFinders.l_gfa1[0], TestUnitLineFinders.gfa1._search_duplicate(TestUnitLineFinders.l_gfa1[0]))
    # path
    self.assertEqual(TestUnitLineFinders.l_gfa1[7], TestUnitLineFinders.gfa1._search_duplicate(TestUnitLineFinders.l_gfa1[7]))

  def test_search_duplicate_gfa2(self):
    # line with mandatory name
    self.assertEqual(TestUnitLineFinders.l_gfa2[0], TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[0]))
    # line with optional name, present
    self.assertEqual(TestUnitLineFinders.l_gfa2[2], TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[2]))
    self.assertEqual(TestUnitLineFinders.l_gfa2[3], TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[3]))
    self.assertEqual(TestUnitLineFinders.l_gfa2[4], TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[4]))
    self.assertEqual(TestUnitLineFinders.l_gfa2[5], TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[5]))
    # line with optional name, not present
    self.assertEqual(None, TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[11]))
    # line with no name
    self.assertEqual(None, TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[6]))
    self.assertEqual(None, TestUnitLineFinders.gfa2._search_duplicate(TestUnitLineFinders.l_gfa2[9]))



import gfapy
import re

def unsafe_decode(string):
  return gfapy.OrientedLine(string[:-1], string[-1])

def decode(string):
  obj = unsafe_decode(string)
  validate_decoded(obj)
  return obj

def validate_encoded(string):
  if not re.match("^[!-~]+[+-]$", string):
    raise gfapy.FormatError(
      "{} is not a valid oriented GFA2 identifier\n".format(repr(string))+
      "(it contains spaces or non-printable characters, or a wrong orientation)")

def validate_decoded(obj):
  if isinstance(obj, gfapy.OrientedLine):
    if not re.match("^[!-~]+$", obj.name):
      raise gfapy.FormatError(
          "{} is not a valid oriented GFA2 identifier\n".format(repr(obj.name)))
    if obj.orient != "+" and obj.orient != "-":
      raise gfapy.FormatError(
          "{} is not a valid orientation\n".format(repr(obj.orient)))
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: gfapy.OrientedLine)")

def unsafe_encode(obj):
  if isinstance(obj, str):
    return obj
  if isinstance(obj, gfapy.OrientedLine):
    return str(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, gfapy.OrientedLine)")

def encode(obj):
  string = unsafe_encode(obj)
  validate_encoded(string)
  return string



import gfapy

class Connection:

  def is_connected(self):
    """
    In a connected line, some of the fields are converted
    into references or a list of references to other lines.
    Furthermore instance variables are populated with back
    references to the line (e.g. connection of a segment
    are stored as references in segment arrays), to allow
    graph traversal.

    Returns
    -------
    bool
      Is the line connected to other lines of a GFA instance?
    """
    return (self._gfa is not None)

  @property
  def gfa(self):
    return self._gfa

  def connect(self, gfa):
    """
    Connect the line to a GFA instance

    Parameters
    ----------
    gfa : GFA
      the GFA instance

    Returns
    -------
    None
    """
    if self.is_connected():
      raise gfapy.RuntimeError(
        "Line {} is already connected to a GFA instance".format(self))
    previous = gfa._search_duplicate(self)
    if previous:
      if previous.virtual:
        return self._substitute_virtual_line(previous)
      else:
        return self._process_not_unique(previous)
    else:
      self._gfa = gfa
      self._initialize_references()
      self._gfa._register_line(self)
      return None

  @property
  def all_references(self):
    """List of lines which contain references to the line instance

    Returns
    -------
    list
    """
    if not self._refs:
      self._refs = {}
    return [x for y in self._refs.values() for x in y]

  def _add_reference(self, line, key, append = True):
    if not self._refs:
      self._refs = {}
    if key not in self._refs:
      self._refs[key] = []
    if append:
      self._refs[key].append(line)
    else:
      self._refs[key].insert(0, line)

  def _initialize_references(self):
    """
    .. note::
      SUBCLASSES with reference fields shall
      overwrite this method to connect their reference
      fields
    """
    if self.REFERENCE_INITIALIZERS:
      for field, klass, refkey in self.REFERENCE_INITIALIZERS:
        self._initialize_reference(field, klass, refkey)

  def _initialize_reference(self, field, klass, refkey):
    name = self.get(field)
    line = self.gfa.line(name)
    if line is None:
      data = [klass.RECORD_TYPE]
      for i in range(len(klass.POSFIELDS)):
        data.append("1")
      line = klass(data, virtual=True, version="gfa2")
      line.name = name
      line.connect(self.gfa)
    self._set_existing_field(field, line, set_reference=True)
    line._add_reference(self, refkey)

  def _process_not_unique(self, previous):
    """
    .. note::
      SUBCLASSES may overwrite this method
      if some kind of non unique lines shall be
      tolerated or handled differently (eg complement links)
    """
    raise gfapy.NotUniqueError(
      "Line: {}\n".format(str(self))+
      "Line or ID not unique\n"+
      "Matching previous line: {}".format(str(previous))
      )



class ToGFA2:

  @property
  def from_coords(self):
    """GFA2 positions of the alignment on the from segment.

    Returns
    -------
    (Integer|Lastpos,Integer|Lastpos)
    	begin and end

    Raises
    ------
    gfapy.ValueError
    	If the overlap is not specified.
    gfapy.RuntimeError
    	If the segment length cannot be determined, because the segment line is unknown.
    gfapy.ValueError
    	If the segment length is not specified in the segment line.
    """
    self._check_overlap()
    if self.from_orient == "+":
      from_l = self._lastpos_of("from")
      return [from_l - self.overlap.length_on_reference(), from_l]
    else:
      return [0, self.overlap.length_on_reference()]

  @property
  def to_coords(self):
    """GFA2 positions of the alignment on the **to** segment."""
    self._check_overlap()
    if self.to_orient == "+":
      return [0, self.overlap.length_on_query()]
    else:
      to_l = self._lastpos_of("to")
      return [to_l - self.overlap.length_on_query(), to_l]



import gfapy
import unittest

class TestAPILinearPaths(unittest.TestCase):

  def test_linear_path_merging(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.1."+"{}".format(sfx))
      with self.assertRaises(gfapy.ValueError):
        gfa.merge_linear_path([["0", "R"],["1", "R"],["2", "L"],["3", "R"]])
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.2."+"{}".format(sfx))
      gfa.merge_linear_path([["0", "R"],["1", "R"],["2", "L"],["3", "R"]])
      with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("0")
      with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("1")
      with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("2")
      with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("3")
      gfa.try_get_segment("0_1_2_3") # nothing raised
      self.assertEqual([], gfa.dovetails)
      self.assertEqual("ACGACGACGTCGA", gfa.segment("0_1_2_3").sequence)

  def test_linear_path_merge_all(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.3."+"{}".format(sfx))
      gfa.merge_linear_paths()
      gfa.merge_linear_paths() # nothing raised
      self.assertEqual(len(gfa.segment_names), 1)
      self.assertIn(gfa.segment_names[0], ["0_1_2_3","3_2_1_0"])
      self.assertEqual(len(gfa.segments), 1)
      self.assertEqual(len(gfa.dovetails), 0)
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.4."+"{}".format(sfx))
      gfa.merge_linear_paths() # nothing raised
      self.assertEqual(3, len(gfa.segments))
      for x in gfa.segments:
        self.assertIn(x.name, {"0","3","1_2","2_1"})
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.5."+"{}".format(sfx))
      gfa.merge_linear_paths() # nothing raised
      self.assertEqual(3, len(gfa.segments))
      self.assertEqual({"0", "1", "2_3"}, {x.name for x in gfa.segments})

  def test_linear_path_merge_example1(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/example1."+"{}".format(sfx))
      lps = set()
      for i, lp in enumerate(gfa.linear_paths()):
        if int(lp[0].name) > int(lp[-1].name):
          lp.reverse()
        lps.add(" ".join([s.name for s in lp]))
      self.assertEqual({"1 19 18", "11 9 12", "22 16 20 21 23"}, lps)

  def test_linear_path_blunt_ends(self):
    for sfx in ["gfa1", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_blunt."+"{}".format(sfx))
      gfa.merge_linear_paths()
      self.assertEqual(1, len(gfa.segments))
      self.assertEqual("s1_s2_s3_s4", gfa.segments[0].name)
      self.assertEqual("CTGAAACGTGGCTCACA", gfa.segments[0].sequence)



import gfapy

class ToGFA2:

  def _to_gfa2_a(self):
    items = []
    for oline in self.captured_path:
      if isinstance(oline.line, gfapy.line.segment.GFA1):
        items.append(str(oline))
      elif isinstance(oline.line, gfapy.line.edge.Link):
        items.append(oline.line.eid + str(oline.orient))
    a = ["O"]
    a.append(self.field_to_s("path_name"))
    a.append(" ".join(items))
    for tn in self.tagnames:
      a.append(self.field_to_s(tn, tag=True))
    return a



import unittest
import gfapy

class TestSequence(unittest.TestCase):
  pass

  #TODO: fix test
  #def test_rc(self):
  #  self.assertEqual("gcatcgatcgt", gfapy.sequence.rc("acgatcgatgc"))
  #  self.assertEqual("gCaTCgatcgt", gfapy.sequence.rc("acgatcGAtGc"))
  #  self.assertEqual("gcatcnatcgt", gfapy.sequence.rc("acgatngatgc"))
  #  self.assertEqual("gcatcYatcgt", gfapy.sequence.rc("acgatRgatgc"))
  #  self.assertRaises(gfapy.InconsistencyError, gfapy.sequence.rc, "acgatUgatgc")
  #  self.assertEqual("gcaucgaucgu", gfapy.sequence.rc("acgaucgaugc"))
  #  self.assertEqual("===.", gfapy.sequence.rc(".==="))
  #  self.assertRaises(gfapy.ValueError, gfapy.sequence.rc, "acgatXgatgc")
  #  self.assertEqual("*", gfapy.sequence.rc("*"))
  #  self.assertRaises(gfapy.ValueError, gfapy.sequence.rc, "**")



import gfapy
from .lines import Lines
from .graph_operations import GraphOperations
from collections import defaultdict
from .rgfa import RGFA
import sys

class Gfa(Lines,GraphOperations,RGFA):
  """Representation of the data in a GFA file.

  Parameters:
    data (str or list): optional, string content of a GFA file, or
       the same as a list (splitted on newlines); default: create an empty Gfa
       instance
    vlevel (int): validation level (default: 1)
    version (str): GFA version ('gfa1' or 'gfa2';
        default: automatic recognition)
    dialect (str): dialect ('standard' or 'rgfa';
        default: standard)

  Raises:
    ~gfapy.error.ArgumentError: if the vlevel or version are invalid
    ~gfapy.error.FormatError: if data is provided, which is invalid
    ~gfapy.error.VersionError: if an unknown version is specified, or data is
        provided, which is not compatible with the specified version
    ~gfapy.error.VersionError: if an unknown dialect is specified
  """

  def __init__(self, *args, vlevel = 1, version = None, dialect = "standard"):
    if not isinstance(vlevel, int):
      raise gfapy.ArgumentError("vlevel is not an integer ({})".format(vlevel))
    if vlevel < 0:
      raise gfapy.ArgumentError(
          "vlevel is not a positive integer ({})".format(vlevel))
    if not version in ['gfa1', 'gfa2', None]:
      raise gfapy.VersionError("GFA version unknown ({})".format(version))
    if not dialect in ['standard', 'rgfa', None]:
      raise gfapy.VersionError("GFA dialect unknown ({})".format(dialect))
    self._vlevel = vlevel
    self._max_int_name = 0
    self._records = defaultdict(dict)
    self._records["H"] = gfapy.line.Header(["H"], vlevel = vlevel)
    self._records["H"].connect(self)
    self._records["S"] = {}
    self._records["P"] = {}
    self._records["F"] = {}
    self._records["\n"] = {}
    self._records["E"] = {}
    self._records["U"] = {}
    self._records["G"] = {}
    self._records["O"] = {}
    self._records["C"] = {}
    self._records["L"] = {}
    self._records["#"] = {}
    self._segments_first_order = False
    self._progress = None
    self._default = {"count_tag": "RC", "unit_length": 1}
    self._line_queue = []
    if version is None:
      self._version = None
      self._version_explanation = None
      self._version_guess = "gfa2"
    else:
      self._version = version
      self._version_explanation = "set during initialization"
      self._version_guess = version
      self._validate_version()
    self._dialect = dialect.lower()
    if len(args) == 1:
      lst = None
      if isinstance(args[0], str):
        lst = args[0].split("\n")
      elif isinstance(args[0], list):
        lst = args[0]
      else:
        raise gfapy.ArgumentError("Cannot create a Gfa"+
            " instance from an object of type {}".format(type(args[0])))
      for line in lst:
          self.add_line(line)
      self.process_line_queue()
      if vlevel >= 1:
        self.validate()
    elif len(args) > 1:
      raise gfapy.ArgumentError("Wrong number of arguments for Gfa()"+
          "({})".format(len(args)))

  @property
  def version(self):
    """GFA version ('gfa1' or 'gfa2')"""
    return self._version

  @version.setter
  def version(self,value):
    self._version=value

  @property
  def dialect(self):
    """GFA dialect ('standard' or 'rgfa')"""
    return self._dialect

  @dialect.setter
  def dialect(self, value):
    self._dialect = value.lower()

  @property
  def vlevel(self):
    """Level of validation"""
    return self._vlevel

  @vlevel.setter
  def vlevel(self,value):
    self._vlevel=value

  def validate(self):
    """Validate the GFA instance

    Checks if all references are solved correctly.
    """
    self.__validate_segment_references()
    self.__validate_path_links()
    self.__validate_group_items()
    self.__validate_gfa2_positions()
    if self._dialect == "rgfa":
      self.validate_rgfa()

  def __str__(self):
    return "\n".join([str(line) for line in self.lines])

  def to_gfa1_s(self):
    """Create a GFA1 string representation for the GFA data

    If the Gfa has version 'gfa1', its string representation is
    returned. Otherwise a conversion from GFA2 is performed.
    """
    if self.version == "gfa1":
      return str(self)
    else:
      lines = []
      for line in self.lines:
        converted = line.to_gfa1_s()
        if converted:
          lines.append(converted)
      return "\n".join(lines)

  def to_gfa1(self):
    """Create a GFA1 Gfa instance for the GFA data

    If the Gfa has version 'gfa1', it is
    returned. Otherwise a conversion from GFA2 is performed.
    """
    if self.version == "gfa1":
      return self
    else:
      gfa1 = gfapy.Gfa(version="gfa1", vlevel=self.vlevel)
      for line in self.lines:
        gfa1.add_line(line.to_gfa1(raise_on_failure=False))
      return gfa1

  def to_gfa2_s(self):
    """Create a GFA2 string representation for the GFA data

    If the Gfa has version 'gfa2', its string representation is
    returned. Otherwise a conversion from GFA1 is performed.
    """
    if self.version == "gfa2":
      return str(self)
    else:
      lines = []
      for line in self.lines:
        converted = line.to_gfa2_s()
        if converted:
          lines.append(converted)
      return "\n".join(lines)

  def to_gfa2(self):
    """Create a GFA2 Gfa instance for the GFA data.

    If the Gfa has version 'gfa2', it is
    returned. Otherwise a conversion from GFA1 is performed.
    """
    if self.version == "gfa2":
      return self
    else:
      gfa2 = gfapy.Gfa(version="gfa2", vlevel=self.vlevel)
      for line in self.lines:
        gfa2.add_line(line.to_gfa2(raise_on_failure=False))
      return gfa2

  # TODO: implement clone (see how clone for lines was implemented)

  def read_file(self, filename):
    """Read GFA data from a file and load it into the Gfa instance.

    Parameters:
      filename (str)
    """
    if self._progress:
      linecount = 0
      with open(filename) as f:
        for line in f:
          linecount += 1
      # TODO: better implementation of linecount
      self._progress_log_init("read_file", "lines", linecount,
                              "Parsing file {}".format(filename)+
                              " containing {} lines".format(linecount))
    with open(filename) as f:
      for line in f:
        self.add_line(line.rstrip('\r\n'))
        if self._progress:
          self._progress_log("read_file")
    if self._line_queue:
      self._version = self._version_guess
      self.process_line_queue()
    if self._progress:
      self._progress_log_end("read_file")
    if self._vlevel >= 1:
      self.validate()
    return self

  @classmethod
  def from_file(cls, filename, vlevel = 1, version = None, dialect="standard"):
    """Create a Gfa instance from the contents of a GFA file.

    Parameters:
      filename (str)
      vlevel (int) : the validation level
      version (str) : the GFA version ('gfa1' or 'gfa2'; default:
          determine version automatically)

    Returns:
      gfapy.Gfa
    """
    gfa = cls(vlevel = vlevel, version = version, dialect = dialect)
    gfa.read_file(filename)
    return gfa

  def to_file(self, filename):
    """Write the content of the instance to a GFA file

    Parameters:
      filename (str)
    """
    with open(filename, "w") as f:
      for line in self.lines:
        f.write(str(line)+"\n")

  def __eq__(self, other):
    self.lines == other.lines

  def __lenstats(self):
    sln = [ s.try_length for s in self.segments ]
    sln = sorted(sln)
    n = len(sln)
    tlen = 0
    for l in sln:
      tlen += l
    n50_target = tlen//2
    n50 = None
    curr_sum = 0
    for l in reversed(sln):
      curr_sum += l
      if curr_sum >= n50_target:
        n50 = l
        break
    q = (sln[0], sln[(n//4)-2], sln[(n//2)-1], sln[((n*3)//4)-1], sln[-1])
    return (q, n50, tlen)

  def __validate_segment_references(self):
    for s in self.segments:
      if s.virtual:
        raise gfapy.NotFoundError("Segment {} ".format(s.name)+
            "does not exist\nReferences to {} ".format(s.name)+
            "were found in the following lines:\n"+s.refstr())

  def __validate_path_links(self):
    for pt in self._gfa1_paths:
      for ol in pt.links:
        l = ol.line
        if l.virtual:
          raise gfapy.NotFoundError("A link equivalent to:\n{}\n".format(\
                  l.to_str(add_virtual_commentary=False))+
              "does not exist, but is required by the following paths:\n"+
              l.refstr())

  def __validate_group_items(self):
    if self.version == "gfa1":
      return
    for group in self.sets + self.paths:
      for item in group.items:
        if isinstance(item, gfapy.OrientedLine):
          item = item.line
        if item.virtual:
          raise gfapy.NotFoundError("A line with identifier {}\n".format(\
                  item.name)+
              "does not exist, but is required by the following groups:\n"+
              item.refstr())

  def __validate_gfa2_positions(self):
    if self.version == "gfa1":
      return
    for line in self.edges + self.fragments:
      line.validate_positions()

  def _validate_version(self):
    if (self._version != None) and (self._version not in gfapy.VERSIONS):
      raise gfapy.VersionError("GFA specification version {} not supported".
              format(self._version))

  # Progress logging related-methods:

  def enable_progress_logging(self, part=0.1, channel=sys.stderr):
    '''Activate logging of progress for some graph operations.

    Parameters:
      part (float) : report when every specified portion of the computation
          is completed (default: 0.1)
      channel : output channel (default: standard error)
    '''
    self._progress = gfapy.Logger(channel=channel)
    self._progress.enable_progress(part=part)

  def _progress_log_init(self, symbol, units, total, initmsg = None):
    if self._progress is not None:
      self._progress.progress_init(symbol, units, total, initmsg)

  def _progress_log(self, symbol, progress=1, **keyargs):
    if self._progress is not None:
      self._progress.progress_log(symbol, progress)

  def _progress_log_end(self, symbol, **keyargs):
    if self._progress is not None:
      self._progress.progress_end(symbol)




import gfapy
import re

def unsafe_decode(string):
  return string.split(" ")

def decode(string):
  validate_encoded(string)
  return unsafe_decode(string)

def validate_encoded(string):
  if not re.match("^[ !-~]+$", string):
    raise gfapy.FormatError(
      "{} is not a valid list of GFA2 identifier\n".format(repr(string))+
      "(it contains non-printable characters)")

def validate_decoded(obj):
  if isinstance(obj, list):
    for elem in obj:
      if isinstance(elem, gfapy.Line):
        elem = str(elem.name)
      elif not isinstance(elem, str):
        raise gfapy.TypeError(
          "the list contains an obj of class {}\n"
          .format(elem.__class__.__name__)+
          "(accepted classes: str, gfapy.Line)")
      if not re.match("^[!-~]+$", elem):
        raise gfapy.FormatError(
        "the list contains an invalid GFA2 identifier ({})\n"
        .format(repr(elem))+
        "(it contains spaces and/or non-printable characters)")
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__) +
      "(accepted classes: list)")

def unsafe_encode(obj):
  if isinstance(obj, list):
    def func(elem):
      if isinstance(elem, str):
        return elem
      elif isinstance(elem, gfapy.Line):
        return str(elem.name)
      else:
        raise gfapy.TypeError(
          "the list contains an obj of class {}\n"
          .format(elem.__class__.__name__)+
          "(accepted classes: str, gfapy.Line)")
    return " ".join(map(func, obj))
  elif isinstance(obj, str):
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: list, str)")

def encode(obj):
  validate_decoded(obj)
  return unsafe_encode(obj)



import gfapy

class Validator:

  @staticmethod
  def _validate_gfa_field(obj, datatype, fieldname = None):
    """Validate the content of a field of a Line instance.

    Parameters:
      obj: the value to be validated. It can be either a string (in which case
        the encoded validation method is used) or any other kind of Python
        object (in which case the decoded validation method is used).
      datatype (str) : the name of the datatype to be used for the validation.
        The datatype name is used for the lookup in the FIELD_MODULE dictiorary
        and the validation method of the returned class is used.
      fieldname (str) : optional, for error messages

    Raises:
      gfapy.error.FormatError : if the format of the string representation is
        invalid; or the object contains strings with an invalid format
      gfapy.error.ValueError : if the value of the decoded field is invalid
      gfapy.error.TypeError : if the specified datatype is not defined or
        if the type of the decoded field is invalid
      gfapy.error.VersionError : if the value is invalid for the GFA version
        for which the datatype is specified
    """
    if isinstance(obj, str):
      Validator.__validate_encoded_gfa_field( obj, datatype, fieldname)
    else:
      Validator.__validate_decoded_gfa_field( obj, datatype, fieldname)

  @staticmethod
  def __validate_decoded_gfa_field(obj, datatype, fieldname = None):
    """Validate a non-string field content.

    Parameters:
      obj : the field content to validate
      datatype (str) : the datatype identifier
      fieldname (str) : for error messages

    Raises:
      gfapy.error.TypeError: if the specified datatype is invalid or the
        object is of a class which is not compatible with the datatype
      gfapy.error.FormatError: if the format of a string in the object
        is not compatible with the datatype; or if the object encoded into
        a GFA string is incompatible with the specification
      gfapy.error.VersionError: if the object value is invalid
        for the specific GFA version for which this datatype is used
      gfapy.error.ValueError: if the value of the object is invalid
    """
    if isinstance(obj, gfapy.FieldArray):
      return obj._validate_gfa_field(datatype, fieldname=fieldname)
    mod = gfapy.Field.FIELD_MODULE.get(datatype)
    if not mod:
      raise gfapy.TypeError(
        "Datatype unknown: {}".format(repr(datatype)))
    return mod.validate_decoded(obj)

  @staticmethod
  def __validate_encoded_gfa_field(obj, datatype, fieldname = None):
    """Validate a string field content.

    Parameters:
      obj (str): the field content to validate
      datatype (str) : the datatype identifier
      fieldname (str) : for error messages

    Raises:
      gfapy.error.TypeError: if the specified datatype is invalid
      gfapy.error.FormatError: if the format of the string is invalid
        for the specified datatype
      gfapy.error.VersionError: if the format of the string is invalid
        for the specific GFA version for which this datatype is used
      gfapy.error.ValueError: if the format of the string is valid,
        but the value encoded by the string is invalid
    """
    mod = gfapy.Field.FIELD_MODULE.get(datatype)
    if not mod:
      raise gfapy.TypeError(
        "Datatype unknown: {}".format(repr(datatype)))
    return mod.validate_encoded(obj)



import gfapy

class Validation:

  def validate_positions(self):
    "Checks that positions suffixed by $ are the last position of segments"
    if self.is_connected():
      for n in ["1","2"]:
        seg = self.get("sid"+n).line
        seq = seg.sequence
        if not gfapy.is_placeholder(seq):
          seqlen = len(seq)
          for pfx in ["beg", "end"]:
            fn = pfx+n
            pos = self.get(fn)
            if gfapy.islastpos(pos):
              if pos != seqlen:
                raise gfapy.InconsistencyError(
                    "Edge: {}\n".format(str(self))+
                    "Field {}: $ after ".format(fn)+
                    "non-last position {}\n".format(str(pos))+
                    "Segment: {}".format(str(seg)))



import gfapy

class LinearPaths:

  def linear_path(self, segment, exclude = None):
    """Find a linear path which contains the specified segment

    Parameters:
      segment (str, Line): the segment to analyse
      exclude : (API private)
    """
    if isinstance(segment, gfapy.Line):
      segment_name = segment.name
    else:
      segment_name = segment
      segment = self.segment(segment_name)
    cs = segment._connectivity()
    if exclude is None:
      exclude = set()
    segpath = gfapy.SegmentEndsPath()
    for i, et in enumerate(["L", "R"]):
      if cs[i] == 1:
        exclude.add(segment_name)
        if len(segpath) > 0:
          segpath.pop()
        segpath += self.__traverse_linear_path(
            gfapy.SegmentEnd(segment, et), exclude)
    return segpath

  def linear_paths(self, redundant_junctions=False):
    """Find linear paths of dovetail overlaps connecting segments.

    Parameters:
      redundant_junctions (bool): output the junction segments at the
        end of each path which involves them; this mimics the construction
        of contigs in string graph assemblers Readjoiner and SGA; default: False
    """
    exclude = set()
    if redundant_junctions:
      junction_exclude = set()
    retval = []
    segnames = self.segment_names
    if self._progress:
      self._progress_log_init("linear_paths", "segments", len(segnames),
          "Detect linear paths ({})".format(len(segnames)))
    for sn in segnames:
      if self._progress:
        self._progress_log("linear_paths")
      if sn in exclude:
        continue
      lp = self.linear_path(sn, exclude)
      if not redundant_junctions:
        if len(lp) > 1:
          retval.append(lp)
      else:
        if lp:
          self._extend_linear_path_to_junctions(lp)
          retval.append(lp)
        else:
          retval += self._junction_junction_paths(sn, junction_exclude)
    if self._progress:
      self._progress_log_end("linear_paths")
    return retval

  def merge_linear_path(self, segpath, redundant_junctions=False, jntag="jn",
                        enable_tracking=False, merged_name=None,
                        cut_counts=False):
    """Merge a specified linear path of dovetail overlaps connecting segments.

    Note:
      for the parameter usage, see merge_linear_paths();
      the only difference is that merged_name can be set to a string (different
      from 'short'), which will be used as a name for the merged segment.
    """
    if len(segpath) < 2:
      return self
    if segpath[0] in [True, False]:
      first_redundant = segpath.pop(0)
      last_redundant = segpath.pop()
    else:
      first_redundant = False
      last_redundant = False
    segpath = [gfapy.SegmentEnd(s) for s in segpath]
    merged, first_reversed, last_reversed = \
        self.__create_merged_segment(segpath,
            redundant_junctions=redundant_junctions, jntag=jntag,
            merged_name=merged_name,cut_counts=cut_counts,
            enable_tracking=enable_tracking)
    self.append(merged)
    if first_redundant:
      self._link_duplicated_first(merged, self.segment(segpath[0].segment),
                                   first_reversed, jntag)
    else:
      self.__link_merged(merged.name, segpath[0].inverted(), first_reversed)
    if last_redundant:
      self._link_duplicated_last(merged, self.segment(segpath[-1].segment),
                                  last_reversed, jntag)
    else:
      self.__link_merged(merged.name, segpath[-1], last_reversed)
    idx1 = 1 if first_redundant else 0
    idx2 = -1 if last_redundant else None
    for sn_et in segpath[idx1:idx2]:
      self.segment(sn_et.segment).disconnect()
      if self._progress:
        self._progress_log("merge_linear_paths", 0.05)
    return self

  def merge_linear_paths(self, redundant_junctions=False, jntag="jn",
                        merged_name=None, enable_tracking=False,
                        cut_counts=False):
    """Find and merge linear paths of dovetail overlaps connecting segments.

    Note:
      Besides obviously the dovetail overlaps, all lines refererring to the
      merged segments (containments, internal edges, paths, sets, fragments,
      gaps) are removed from the Gfa instance.

    Parameters:
      merged_name (str): if 'short', then a name is computed using an unused
        integer; otherwise the name is computed using a combination of the
        names of the merged segments, separated by an underscore
      cut_counts (bool): if True, the total count in merged segment m,
         composed of segments s of set S is multiplied by the factor
         ``Sum(|s in S|)/|m|``
      enable_tracking: if True, tracking information is added as follows;
        the name of the component segments is stored in the ``or`` tag (or the
        content of their ``or`` tag, instead of the name, if any) and their
        starting positions is stored in the ``mp`` tag; the ``rn`` tag, used
        for storing possibe inversion positions by the random orientation
        methods of this library, is inherited and the positions updated;
        unless merged_name is set to 'short', the computation of the merged
        name is enhanced, in that reverse complement components are suffixed
        with ``^`` and parenthesis added by the random orientation methods of
        this library are inherited
      redundant_junctions (bool): output the junction segments at the
        end of each path which involves them; this mimics the construction
        of contigs in string graph assemblers Readjoiner and SGA; default: False
      jntag (str) : the tag to use for the temporary storage of junction
        information, if the redundant_junctions flag is set (default: jn)
    """
    paths = self.linear_paths(redundant_junctions)
    if self._progress:
      psize = sum([len(path) for path in paths])
      self._progress_log_init("merge_linear_paths", "segments", psize,
          "Merge {} linear paths ".format(len(paths))+
          "({} segments)".format(psize))
    for path in paths:
      self.merge_linear_path(path, redundant_junctions=redundant_junctions,
                             jntag=jntag, merged_name=merged_name,
                             cut_counts=cut_counts,
                             enable_tracking=enable_tracking)
    if self._progress:
      self._progress_log_end("merge_linear_paths")
    if redundant_junctions:
      self._remove_junctions(jntag)
    return self

  def __traverse_linear_path(self, segment_end, exclude):
    lst = gfapy.SegmentEndsPath()
    current = gfapy.SegmentEnd(segment_end)
    current.segment = self.segment(current.segment)
    while True:
      after = current.segment.dovetails_of_end(current.end_type)
      before = current.segment.dovetails_of_end(gfapy.invert(current.end_type))
      if (len(before) == 1 and len(after) == 1) or not lst:
        lst.append(gfapy.SegmentEnd(current.name, current.end_type))
        exclude.add(current.name)
        current = after[0].other_end(current).inverted()
        if current.name in exclude:
          break
      elif len(before) == 1:
        lst.append(gfapy.SegmentEnd(current.name, current.end_type))
        exclude.add(current.name)
        break
      else:
        break
    if segment_end.end_type == "L":
      return list(reversed(lst))
    else:
      return lst

  def __sum_of_counts(self, segpath, multfactor = 1):
    retval = {}
    segs = [self.try_get_segment(sn_et.segment) for sn_et in segpath]
    for count_tag in ["KC","RC","FC"]:
      for s in segs:
        if count_tag in s.tagnames:
          if count_tag not in retval:
            retval[count_tag] = 0
          retval[count_tag] += int(retval[count_tag]*multfactor)
    return retval

  def _add_segment_to_merged(self, merged, segment, is_reversed, cut, init,
                             enable_tracking=False, merged_name=None):
    n = segment.name
    if is_reversed:
      s = gfapy.sequence.rc(segment.sequence)[cut:]
      if enable_tracking:
        n = self._reverse_segment_name(segment.name, "_")
        rn = self._reverse_pos_array(segment.rn, segment.LN)
        mp = self._reverse_pos_array(segment.mp, segment.LN)
    else:
      s = segment.sequence[cut:]
      if enable_tracking:
        rn = segment.rn
        mp = segment.mp
    if enable_tracking:
      if not mp and segment.LN:
        mp = [1, segment.LN]
      if segment.get("or") is None:
        o = n
      elif is_reversed:
        o = self._reverse_segment_name(segment.get("or"), ",")
      else:
        o = segment.get("or")
    if init:
      merged.sequence = [s]
      if merged_name:
        merged.name = [merged_name]
      else:
        merged.name = [n]
      merged.LN = segment.LN
      if enable_tracking:
        merged.rn = rn
        merged.set("or",[o])
        merged.mp = mp
    else:
      if gfapy.is_placeholder(segment.sequence):
        merged.sequence = gfapy.Placeholder()
      else:
        merged.sequence.append(s)
      if not merged_name:
        merged.name.append(n)
      if merged.LN:
        if enable_tracking:
          if rn:
            rn = [pos - cut + merged.LN for pos in rn]
            if not merged.rn:
              merged.rn = rn
            else:
              merged.rn += rn
          if mp and merged.mp:
            merged.mp += [pos - cut + merged.LN for pos in mp]
        if segment.LN:
          merged.LN += (segment.LN - cut)
        else:
          merged.LN = None
      elif enable_tracking:
        merged.mp = None
      if enable_tracking:
        if not merged.get("or"):
          merged.set("or", [o])
        else:
          merged.get("or").append(o)

  @staticmethod
  def _reverse_segment_name(name, separator):
    retval = []
    for part in name.split(separator):
      has_openp = part[0] == "("
      has_closep = part[-1] == ")"
      if has_openp:
        part = part[1:-2]
      if has_closep:
        part = part[:-1]
      if part[-1] == "^":
        part = part[:-1]
      else:
        part+="^"
      if has_openp:
        part+=")"
      if has_closep:
        part+="("+part
      retval.append(part)
    return separator.join(reversed(retval))

  @staticmethod
  def _reverse_pos_array(pos_array, lastpos):
    if pos_array is None or lastpos is None:
      return None
    else:
      return [lastpos-pos+1 for pos in pos_array].reverse()

  def __create_merged_segment(self, segpath, redundant_junctions=False,
      jntag="jn", merged_name=None, enable_tracking=False, cut_counts=False):
    merged = self.try_get_segment(segpath[0].segment).clone()
    merged.set(jntag, None)
    merged_vlevel = merged.vlevel
    merged.vlevel = 0
    total_cut = 0
    a = segpath[0]
    first_reversed = (a.end_type == "L")
    last_reversed = None
    if merged_name == "short":
      merged_name = self.unused_name()
    self._add_segment_to_merged(merged, self.segment(a.segment),
        first_reversed, 0, True, enable_tracking=enable_tracking,
        merged_name=merged_name)
    if self._progress:
      self._progress_log("merge_linear_paths", 0.95)
    for i in range(len(segpath)-1):
      b = gfapy.SegmentEnd(segpath[i+1]).inverted()
      ls = self.segment(a.segment).end_relations(a.end_type, b, "dovetails")
      if len(ls) != 1:
        msg = "A single link was expected between {}".format(a) + \
              "and {}".format(b) + "{} were found".format(len(ls))
        raise gfapy.ValueError(msg)
      l = ls[0]
      if not l.overlap:
        cut = 0
      elif all(op.code in ["M","="] for op in l.overlap):
        cut = sum([len(op) for op in l.overlap])
      else:
        raise gfapy.ValueError(
            "Merging is only allowed if all operations are M/=")
      total_cut += cut
      last_reversed = (b.end_type == "R")
      self._add_segment_to_merged(merged, self.segment(b.segment),
          last_reversed, cut, False, enable_tracking=enable_tracking,
          merged_name=merged_name)
      a = gfapy.SegmentEnd(b).inverted()
      if self._progress:
        self._progress_log("merge_linear_paths", 0.95)
    merged.vlevel = merged_vlevel
    if isinstance(merged.name, list):
      merged.name = "_".join(merged.name)
    ortag = merged.get("or")
    if isinstance(ortag, list):
      merged.set("or", ",".join(ortag))
    if not gfapy.is_placeholder(merged.sequence):
      merged.sequence = "".join(merged.sequence)
      if self._version == "gfa1":
        if not merged.LN:
          merged.LN = len(merged.sequence)
        elif self._vlevel > 0 and merged.LN != len(merged.sequence):
          raise gfapy.InconsistencyError(
              "Computed sequence length {} ".format(len(merged.sequence))+
              "and computed LN {} differ".format(merged.LN))
    if merged.length is not None:
      for count_tag in ["KC", "RC", "FC"]:
        merged.set(count_tag, None)
    else:
      factor = 1
      if cut_counts:
        factor = merged.length / (total_cut+merged.length)
      for count_tag,count in self.__sum_of_counts(segpath,factor).items():
        merged.set(count_tag, count)
    return merged, first_reversed, last_reversed

  def __link_merged(self, merged_name, segment_end, is_reversed):
    for l in self.segment(segment_end.segment).dovetails_of_end(
                                                 segment_end.end_type):
      l2 = l.clone()
      if l2.to_segment == segment_end.segment:
        l2.to_segment = merged_name
        if is_reversed:
          l2.to_orient = gfapy.invert(l2.to_orient)
      else:
        l2.from_segment = merged_name
        if is_reversed:
          l2.from_orient = gfapy.invert(l2.from_orient)
      l.disconnect()
      self.add_line(l2)




import gfapy

class ToGFA2:
  """
  Access of / conversion from a GFA1 link/containment as / to a GFA2 edge.
  """

  @property
  def eid(self):
    """The content of the ID tag"""
    i = self.get("ID")
    if i is None:
      return gfapy.Placeholder()
    return i

  name = eid

  @property
  def sid1(self):
    """The combination of the from_segment and from_orientation fields"""
    return self.oriented_from

  @property
  def sid2(self):
    """The combination of the to_segment and to_orientation fields"""
    return self.oriented_to

  @property
  def beg1(self):
    """The start coordinate of the alignment on the from segment"""
    return self.from_coords[0]

  @property
  def end1(self):
    """The end coordinate of the alignment on the from segment"""
    return self.from_coords[1]

  @property
  def beg2(self):
    """The start coordinate of the alignment on the to segment"""
    return self.to_coords[1]

  @property
  def end2(self):
    """The end coordinate of the alignment on the to segment"""
    return self.to_coords[1]

  @property
  def alignment(self):
    """The content of the overlap field (CIGAR or Placeholder)"""
    return self.overlap

  def _to_gfa2_a(self):
    a = ["E"]
    if not self.get("ID") and self.is_connected():
      self.set("ID", self._gfa.unused_name())
    if self.get("ID"):
      a.append(str(self.get("ID")))
    else:
      a.append("*")
    a.append(str(self.sid1))
    a.append(str(self.sid2))
    a += [ str(x) for x in self.from_coords ]
    a += [ str(x) for x in self.to_coords ]
    try:
      self.overlap.validate(version = "gfa2")
    except:
      raise gfapy.RuntimeError(
        "Conversion of edge line from GFA1 to GFA2 failed\n"+
        "Overlap is invalid or not compatible with GFA2\n"+
        "Edge line: {}\n".format(str(self)))
    a.append(self.field_to_s("overlap"))
    for fn in self.tagnames:
      if fn != "ID":
        a.append(self.field_to_s(fn, tag = True))
    return a

  def _lastpos_of(self, field):
    line = getattr(self,field)
    if not isinstance(line, gfapy.Line):
      raise gfapy.RuntimeError(
        "Line {} is not embedded in a GFA object".format(self))
    length = line.length
    if length is None:
      raise gfapy.ValueError(
        "Length of segment {} unknown".format(self.to.name))
    return gfapy.LastPos(length)

  def _check_overlap(self):
    if isinstance(self.overlap, gfapy.Placeholder):
      raise gfapy.ValueError(
        "Link: {}\n".format(self)+
        "Missing overlap, cannot compute overlap coordinates")



class DefaultRecordDefinition:
  """
  Provides default values for the constants for the definition of record types.
  """
  RECORD_TYPE = None
  POSFIELDS = []
  PREDEFINED_TAGS = []
  DATATYPE = {}
  NAME_FIELD = None
  STORAGE_KEY = None
  FIELD_ALIAS = None
  REFERENCE_FIELDS = []
  BACKREFERENCE_RELATED_FIELDS = []
  DEPENDENT_LINES = []
  OTHER_REFERENCES = []
  REFERENCE_INITIALIZERS = None




import gfapy
import unittest

class TestAPIGroupsValidation(unittest.TestCase):

  def test_invalid_path_gfa2(self):
    with self.assertRaises(gfapy.NotFoundError):
      g = gfapy.Gfa.from_file("tests/testdata/invalid_path.gfa2")

  def test_invalid_path_gfa2_vlevel0(self):
    g = gfapy.Gfa.from_file("tests/testdata/invalid_path.gfa2", vlevel = 0)
    with self.assertRaises(gfapy.NotFoundError):
      g.validate()

  def test_valid_path_gfa2(self):
    # nothing raised
    g = gfapy.Gfa.from_file("tests/testdata/valid_path.gfa2")



#!/usr/bin/env python3
"""
Run the benchmarks necessary to reproduce the figures of Section 3
of the Supplementary Information of the manuscript \"Gfapy: a flexible
and extensible software library for handling sequence graphs in Python\"
and plots the figures using R.
"""

import argparse
import os

op = argparse.ArgumentParser(description=__doc__)
op.add_argument("fignum", help="Figure number", type=int,
    choices=range(5,9))
op.add_argument("--queue", default=None,
    help="Use the specified queue of a Grid Engine cluster system "+
    "(e.g. 16c.q). If not provided, the benchmarks are run on the "+
    "local computer.")
op.add_argument("--nrepl",type=int, default=3,
    help="Number of replicates (default: 3)")
op.add_argument("--fast",action="store_true",
    help="Run only the three fastest datapoints of the benchmark")
opts = op.parse_args()

if opts.fignum == 5:
  testvar="segments"
  operation="convert"
elif opts.fignum == 6:
  testvar="connectivity"
  operation="convert"
elif opts.fignum == 7:
  testvar="segments"
  operation="mergelinear"
else: # 8
  testvar="connectivity"
  operation="mergelinear"

if opts.fast:
  subset="fast"
else:
  subset="all"

run_benchmarks_args="figure{}.out {} gfa2 {} {} {}".format(
    opts.fignum, operation, testvar, subset, opts.nrepl)

if not opts.queue:
  os.system("./gfapy-run-benchmarks.sh {}".format(run_benchmarks_args))
else:
  qsub_script_pfx=\
"""#!/bin/bash
#$ -clear
#$ -q {}
#$ -cwd
#$ -V
#$ -S /bin/bash
#$ -o jobs_out
#$ -j y
#$ -sync y

""".format(opts.queue)
  with open("gfapy-run-benchmarks.sh", "r") as input_file:
    content = input_file.read()
  with open("gfapy-run-benchmarks.qsub", "w") as output_file:
    output_file.write(qsub_script_pfx)
    output_file.write(content)
  os.system("mkdir -p jobs_out")
  os.system("qsub gfapy-run-benchmarks.qsub {}".format(run_benchmarks_args))

if testvar == "segments":
  prepareflag=""
else:
  prepareflag="--mult"
os.system("./gfapy-plot-preparedata.py {} figure{}.out > figure{}.dat".format(
  prepareflag, opts.fignum, opts.fignum))
os.system("./gfapy-plot-benchmarkdata.R figure{}.dat figure{} {}".format(
  opts.fignum, opts.fignum, testvar))



import gfapy

class Placeholder:
  """
  A placeholder is used in mandatory fields when a value is not specified.

  Its string representation is an asterix ``*``.
  """

  def __str__(self):
    return "*"

  def __repr__(self):
    return "gfapy.Placeholder()"

  def __bool__(self):
    return False

  def complement(self):
    return self

  def is_empty(self):
    """
    A placeholder is always empty.

    Returns
    -------
    True : bool
      Returns always **True**
    """
    return True

  def validate(self, *args, **keyargs):
    """
    A placeholder is always valid.
    """
    return None

  def rc(self):
    """
    For compatibility with String#rc (gfapy.sequence module).

    Returns
    -------
    self : Placeholder
    """
    return self

  def __len__(self):
    """
    Length/size of a placeholder is always 0.

    Returns
    -------
    0 : int
      Returns always 0
    """
    return 0

  def __getitem__(self, key):
    """
    Any cut of the placeholder returns the placeholder itself.

    Parameters
    ----------
    key
      Ignored

    Returns
    -------
    self : Placeholder
    """
    return self

  def __add__(self, other):
    """
    Adding the placeholder to anything returns the placeholder itself.

    Parameters
    ----------
    other
      Ignored

    Returns
    -------
    self : Placeholder
    """
    return self

  def __eq__(self, other):
    return gfapy.is_placeholder(other)

def is_placeholder(object):
  if object is Placeholder:
    return True
  elif object is None:
    return True
  elif object == "*":
    return True
  elif isinstance(object, list) and len(object) == 0:
    return True
  else:
    return False



import gfapy

def unsafe_decode(string):
  return string

def decode(string):
  validate_encoded(string)
  return string

def validate_encoded(string):
  if string.find("\n") != -1:
    raise gfapy.FormatError("{} is not a single-line string"
                            .format(repr(string)))

validate_decoded = validate_encoded

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(object.__class__.__name__)+
      "(accepted classes: str)")



class Artifacts:

  def remove_small_components(self, minlen):
    """Remove connected components with combined segment length < minlen.

    Note:
      Connected components of the graph are computed, considering only dovetail
      overlaps as connection of segments.

    Parameters:
      minlen (int) : the minimal length of the components to keep.
    """
    for cc in filter(lambda c: sum([self.segment(sn).length for sn in c]) \
                     < minlen, self.connected_components()):
      for s in cc:
        self.rm(s)

  def remove_dead_ends(self, minlen):
    """Remove dead end segments from the graph.

    Dead end segments are defined as segment, with a sequence smaller
    than a given minlen parameter, and whose removal does not split
    connected components in the graph, and which have no connections (dovetail
    overlaps) for at least one of the two ends of the sequence.

    Note:
      Only dovetail overlaps are considered as connections.

    Parameters:
      minlen (int) : the minimal length of an end to keep.
    """
    for s in self.segments:
      c = s._connectivity()
      if s.length < minlen and \
        (c[0]==0 or c[1]==0) and \
          not self.is_cut_segment(s):
        self.rm(s)



import gfapy

class ToGFA1:

  def _to_gfa1_a(self):
    a = ["P"]
    if gfapy.is_placeholder(self.name):
      raise gfapy.ValueError(
        "Conversion to GFA1 failed\n"+
        "The path name is a placeholder\t"+
        "Line: {}".format(self))
    a.append(self.name)
    segment_names = []
    for oline in self.captured_segments:
      gfapy.Field._validate_gfa_field(oline.name, "segment_name_gfa1")
      segment_names.append(str(oline))
    a.append(",".join(segment_names))
    overlaps = []
    for oline in self.captured_edges:
      gfapy.Field._validate_gfa_field(oline.line.overlap, "alignment_gfa1")
      overlaps.append(str(oline.line.overlap))
    a.append(",".join(overlaps))
    for tn in self.tagnames:
      a.append(self.field_to_s(tn, tag=True))
    return a



import gfapy
import re

def decode(string):
  try:
    return float(string)
  except:
    raise gfapy.FormatError

unsafe_decode = decode

def validate_decoded(integer):
  pass
  # always valid

def validate_encoded(string):
  if not re.match(r"^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$", string):
    raise gfapy.FormatError(
      "{} does not represent a valid float\n".format(repr(string)) +
      "(it does not match [-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)")

def unsafe_encode(obj):
  return str(obj)


def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, int) or isinstance(obj, float):
    return str(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, int, float)")



class WriterWoSequence:

  def __str__(self, without_sequence = False):
    """
    Parameters
    ----------
    without_sequence : bool
    	 If **True**, output "*" instead of sequence.
    
    Returns
    -------
    str
    	String representation of the segment.
    """
    if not without_sequence:
      return super().__str__()
    else:
      saved = self.sequence
      self.sequence = "*"
      retval = super().__str__()
      self.sequence = saved
      return retval



import gfapy

class LengthGFA1:

  @property
  def length(self):
    """
    Returns
    -------
    int
      Value of LN tag, if segment has LN tag.
    int
      Sequence length if no LN and sequence not "*".
    None
      If sequence is "*".

    See Also
    --------
    try_get_length
    """
    if self.LN:
      return self.LN
    elif not gfapy.is_placeholder(self.sequence):
      return len(self.sequence)
    else:
      return None

  def try_get_length(self):
    """
    Raises
    ------
    gfapy.NotFoundError
      If not an LN tag and the sequence is "*".

    See Also
    --------
    __len__
    """
    l = self.length
    if l is None:
      raise gfapy.NotFoundError("No length information available")
    return l

  def validate_length(self):
    """
    Raises
    ------
    gfapy.InconsistencyError
      If sequence length and LN tag are not consistent.
    """
    if not gfapy.is_placeholder(self.sequence) and "LN" in self.tagnames:
      if self.LN != len(self.sequence):
        raise gfapy.InconsistencyError(
          "Segment: {}\n".format(str(self))+
          "Length in LN tag ({}) ".format(self.LN)+
          "is different from length of sequence field ({})"
          .format(len(self.sequence)))

  def _validate_record_type_specific_info(self):
    self.validate_length()



from ..common.from_to import FromTo
from ..common.alignment_type import AlignmentType
from ..gfa1.to_gfa2 import ToGFA2 as GFA1_ToGFA2
from ..gfa1.alignment_type import AlignmentType as GFA1_AlignmentType
from ..gfa1.oriented_segments import OrientedSegments
from ..gfa1.references import References
from ..gfa1.other import Other
from ..containment.canonical import Canonical
from ..containment.pos import Pos
from ..containment.to_gfa2 import ToGFA2 as Containment_ToGFA2
from ..edge import Edge

class Containment(Containment_ToGFA2, Pos, Canonical, Other,
                  GFA1_AlignmentType, OrientedSegments, References,
                  GFA1_ToGFA2, AlignmentType, FromTo, Edge):
  """A containment line (C) of a GFA1 file

  Note:
    from_segment and to_segment are used instead of from/to
    as from is not a valid method name in Python. However, when not
    used as method name (e.g. as argument of get()), from and to can
    be used, as an alias has been defined.

  Note:
    The from segment is considered the container, the to segment the contained
    sequence. This is not indicated in the specification, but examples where
    done with this assumption in the GFA forum.
  """
  RECORD_TYPE = "C"
  POSFIELDS = ["from_segment", "from_orient", "to_segment",
               "to_orient", "pos", "overlap"]
  FIELD_ALIAS = {"container" : "from_segment",
                 "contained" : "to_segment",
                 "from" : "from_segment",
                 "to" : "to_segment",
                 "container_orient" : "from_orient",
                 "contained_orient" : "to_orient"}
  PREDEFINED_TAGS = ["MQ", "NM", "ID"]
  NAME_FIELD = "ID"
  DATATYPE = {
     "from_segment" : "segment_name_gfa1",
     "from_orient" : "orientation",
     "to_segment" : "segment_name_gfa1",
     "to_orient" : "orientation",
     "pos" : "position_gfa1",
     "overlap" : "alignment_gfa1",
     "MQ" : "i",
     "NM" : "i",
     "ID" : "Z",
  }
  REFERENCE_FIELDS = ["from_segment", "to_segment"]

Containment._apply_definitions()



import gfapy
import unittest

class TestAPIReferencesEdgesGFA2(unittest.TestCase):

  def test_edges_references(self):
    g = gfapy.Gfa()
    lab = gfapy.Line("E\t*\ta+\tb+\t0\t10\t90\t100$\t*")
    self.assertEqual(gfapy.OrientedLine("a","+"), lab.sid1)
    self.assertEqual(gfapy.OrientedLine("b","+"), lab.sid2)
    sa = gfapy.Line("S\ta\t100\t*")
    g.append(sa)
    sb = gfapy.Line("S\tb\t100\t*")
    g.append(sb)
    g.append(lab)
    self.assertEqual(sa, lab.sid1.line)
    self.assertEqual(sb, lab.sid2.line)
    lab.disconnect()
    self.assertEqual("a", lab.sid1.line)
    self.assertEqual("b", lab.sid2.line)
    # disconnection of segment cascades on edges
    g.append(lab)
    assert(lab.is_connected())
    self.assertEqual(sa, lab.sid1.line)
    sa.disconnect()
    assert(not lab.is_connected())
    self.assertEqual("a", lab.sid1.line)

  def test_edges_backreferences(self):
    g = gfapy.Gfa()
    sa = gfapy.Line("S\ta\t100\t*")
    g.append(sa)
    s = {}
    for sbeg1, beg1 in {"0":0,"1":30,"2":70,"$":gfapy.LastPos("100$")}.items():
      for send1, end1 in {"0":0,"1":30,"2":70,"$":gfapy.LastPos("100$")}.items():
        if beg1 > end1:
          continue
        for sbeg2, beg2 in {"0":0,"1":30,"2":70,"$":gfapy.LastPos("100$")}.items():
          for send2, end2 in {"0":0,"1":30,"2":70,"$":gfapy.LastPos("100$")}.items():
            if beg2 > end2:
              continue
            for or1 in ["+","-"]:
              for or2 in ["+","-"]:
                eid = "<{}".format(or1)+"{}".format(or2)+"{}".format(sbeg1)+"{}".format(send1)+"{}".format(sbeg2)+"{}".format(send2)
                other = "s{}".format(eid)
                g.append("\t".join(["E",eid,"a{}".format(or1),"{}".format(other)+"{}".format(or2),
                                    str(beg1),str(end1),str(beg2),str(end2),"*"]))
                s[other] = gfapy.Line("S\t{}".format(other)+"\t100\t*")
                g.append(s[other])
                eid = ">{}".format(or1)+"{}".format(or2)+"{}".format(sbeg1)+"{}".format(send1)+"{}".format(sbeg2)+"{}".format(send2)
                other = "s{}".format(eid)
                g.append("\t".join(["E",eid,"{}".format(other)+"{}".format(or1),"a{}".format(or2),
                                    str(beg1),str(end1),str(beg2),str(end2),"*"]))
                s[other] = gfapy.Line("S\t{}".format(other)+"\t100\t*")
                g.append(s[other])
    exp_sa_d_L = []
    exp_sa_d_R = []
    exp_sa_e_cr = []
    exp_sa_e_cd = []
    exp_sa_i = []
    # a from 0 to non-$, other from non-0 to $;
    # same orientation;"d_L"
    # opposite orientations;"internals"
    for e_a in ["0","1","2"]:
      for b_other in ["1","2","$"]:
        for ors in ["++","--"]:
          exp_sa_d_L.append("<{}".format(ors)+"0{}".format(e_a)+"{}".format(b_other)+"$")
          exp_sa_d_L.append(">{}".format(ors)+"{}".format(b_other)+"$0{}".format(e_a))
        for ors in ["+-","-+"]:
          exp_sa_i.append("<{}".format(ors)+"0{}".format(e_a)+"{}".format(b_other)+"$")
          exp_sa_i.append(">{}".format(ors)+"{}".format(b_other)+"$0{}".format(e_a))
    # one from non-0 to non-$, other non-0 to non-$;"internals"
    for pos_one in ["11","12","22"]:
      for pos_other in ["11","12","22"]:
        for ors in ["++","--","+-","-+"]:
          for d in ["<",">"]:
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos_one)+"{}".format(pos_other))
    # one from non-0 to non-$, other 0 to non-$;"internals"
    for pos_one in ["11","12","22"]:
      for pos_other in ["00","01","02"]:
        for ors in ["++","--","+-","-+"]:
          for d in ["<",">"]:
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos_one)+"{}".format(pos_other))
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos_other)+"{}".format(pos_one))
    # one from non-0 to non-$, other non-0 to $;"internals"
    for pos_one in ["11","12","22"]:
      for pos_other in ["1$","2$","$$"]:
        for ors in ["++","--","+-","-+"]:
          for d in ["<",">"]:
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos_one)+"{}".format(pos_other))
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos_other)+"{}".format(pos_one))
    # other from 0 to non-$, a from non-0 to $
    # same orientation;"d_R"
    # opposite orientations;"internals"
    for e_other in ["0","1","2"]:
      for b_a in ["1","2","$"]:
        for ors in ["++","--"]:
          exp_sa_d_R.append("<"+"{}".format(ors)+"{}".format(b_a)+"$0"+"{}".format(e_other))
          exp_sa_d_R.append(">"+"{}".format(ors)+"0"+"{}".format(e_other)+"{}".format(b_a)+"$")
        for ors in ["+-","-+"]:
          exp_sa_i.append("<"+"{}".format(ors)+"{}".format(b_a)+"$0"+"{}".format(e_other))
          exp_sa_i.append(">"+"{}".format(ors)+"0"+"{}".format(e_other)+"{}".format(b_a)+"$")
    # both from 0 to non-$,
    # opposite orientations;"d_L"
    # same orientation;"internals"
    for e1 in ["0","1","2"]:
      for e2 in ["0","1","2"]:
        pos = "0"+"{}".format(e1)+"0"+"{}".format(e2)
        for ors in ["+-","-+"]:
          for d in ["<",">"]:
            exp_sa_d_L.append("{}".format(d)+"{}".format(ors)+"{}".format(pos))
        for ors in ["++","--"]:
          for d in ["<",">"]:
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos))
    # both from non-0 to $,
    # opposite orientations;"d_R"
    # same orientation;"internals"
    for e1 in ["1","2","$"]:
      for e2 in ["1","2","$"]:
        pos = "{}".format(e1)+"$"+"{}".format(e2)+"$"
        for ors in ["+-","-+"]:
          for d in ["<",">"]:
            exp_sa_d_R.append("{}".format(d)+"{}".format(ors)+"{}".format(pos))
        for ors in ["++","--"]:
          for d in ["<",">"]:
            exp_sa_i.append("{}".format(d)+"{}".format(ors)+"{}".format(pos))
    # a whole; other non-whole:edges_to_containers
    for pos_other in ["00","01","02","11","12","1$","22","2$","$$"]:
      for ors in ["++","--","+-","-+"]:
        exp_sa_e_cr.append("<{}".format(ors)+"0${}".format(pos_other))
        exp_sa_e_cr.append(">{}".format(ors)+"{}".format(pos_other)+"0$")
    # a not-whole; other whole:edges_to_contained
    for pos_a in ["00","01","02","11","12","1$","22","2$","$$"]:
      for ors in ["++","--","+-","-+"]:
        exp_sa_e_cd.append("<{}".format(ors)+"{}".format(pos_a)+"0$")
        exp_sa_e_cd.append(">{}".format(ors)+"0${}".format(pos_a))
    # a sid1; both whole:edges_to_contained
    for ors in ["++","--","+-","-+"]:
      exp_sa_e_cd.append("<{}".format(ors)+"0$0$")
    # a sid2; both whole:edges_to_containers
    for ors in ["++","--","+-","-+"]:
      exp_sa_e_cr.append(">{}".format(ors)+"0$0$")
    # dovetails_[LR]
    self.assertEqual(set(exp_sa_d_L), set([x.name for x in sa.dovetails_L]))
    self.assertEqual(set(exp_sa_d_R), set([x.name for x in sa.dovetails_R]))
    # dovetails()
    self.assertEqual(sa.dovetails_L, sa.dovetails_of_end("L"))
    self.assertEqual(sa.dovetails_R, sa.dovetails_of_end("R"))
    self.assertEqual((sa.dovetails_L + sa.dovetails_R), sa.dovetails)
    # neighbours
    self.assertEqual(set(["s"+x for x in (exp_sa_d_L+exp_sa_d_R)]),
                     set([x.name for x in sa.neighbours]))
    # edges_to_containers/contained
    self.assertEqual(set(exp_sa_e_cr),
                     set([x.name for x in sa.edges_to_containers]))
    self.assertEqual(set(exp_sa_e_cd),
                     set([x.name for x in sa.edges_to_contained]))
    # containments
    self.assertEqual(set(exp_sa_e_cr+exp_sa_e_cd),
                 set([x.name for x in sa.containments]))
    # contained/containers
    self.assertEqual(set(["s"+x for x in exp_sa_e_cr]),
                     set([x.name for x in sa.containers]))
    self.assertEqual(set(["s"+x for x in exp_sa_e_cd]),
                     set([x.name for x in sa.contained]))
    # internals
    self.assertEqual(set(exp_sa_i), set([x.name for x in sa.internals]))
    # upon disconnection
    sa.disconnect()
    self.assertEqual([], sa.dovetails_L)
    self.assertEqual([], sa.dovetails_R)
    self.assertEqual([], sa.dovetails_of_end("L"))
    self.assertEqual([], sa.dovetails_of_end("R"))
    self.assertEqual([], sa.neighbours)
    self.assertEqual([], sa.edges_to_containers)
    self.assertEqual([], sa.edges_to_contained)
    self.assertEqual([], sa.containments)
    self.assertEqual([], sa.contained)
    self.assertEqual([], sa.containers)
    self.assertEqual([], sa.internals)




class Topology:

  def is_circular(self):
    """
    Is the path circular? In this case the number of CIGARs must be
    equal to the number of segments.

    Returns
    -------
    bool
    """
    return len(self.overlaps) == len(self.segment_names)

  def is_linear(self):
    """
    Is the path linear? This is the case when the number of CIGARs
    is equal to the number of segments minus 1, or the CIGARs are
    represented by a single "*".
    """
    return not self.is_circular()



import unittest
import gfapy

class TestInternalsTagDatatype(unittest.TestCase):

  def test_datatype_value_independent(self):
    self.assertEqual("Z", gfapy.Field._get_default_gfa_tag_datatype("string"))
    self.assertEqual("i", gfapy.Field._get_default_gfa_tag_datatype(1))
    self.assertEqual("f", gfapy.Field._get_default_gfa_tag_datatype(1.0))
    self.assertEqual("H", gfapy.Field._get_default_gfa_tag_datatype(gfapy.ByteArray([])))
    self.assertEqual("B", gfapy.Field._get_default_gfa_tag_datatype(gfapy.NumericArray([])))
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype({}))

  def test_datatype_arrays(self):
    self.assertEqual("B", gfapy.Field._get_default_gfa_tag_datatype([1,1]))
    self.assertEqual("B", gfapy.Field._get_default_gfa_tag_datatype([1.0,1.0]))
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype([1,1.0]))
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype(["1",1]))
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype([1.0,"1.0"]))
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype(["z","z"]))
    self.assertEqual("J", gfapy.Field._get_default_gfa_tag_datatype(
                                        [[1,2,3],[3,4,5]]))



import gfapy

class InducedSet:

  @property
  def induced_set(self):
    if not self.is_connected:
      raise gfapy.RuntimeError(
        "Induced set cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    iss = self.induced_segments_set
    ise = self._compute_induced_edges_set(iss)
    return iss + ise

  @property
  def induced_edges_set(self):
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Induced set cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    return self._compute_induced_edges_set(self.induced_segments_set)

  @property
  def induced_segments_set(self):
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Induced set cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    segments_set = list()
    for item in self.items:
      if isinstance(item, str):
        raise gfapy.RuntimeError(
          "Induced set cannot be computed; a reference has not been resolved\n"+
          "Line: {}\n".format(self)+
          "Unresolved reference: {} (String found)".format(item.line))
      elif isinstance(item, gfapy.line.segment.GFA2):
        self._check_induced_set_elem_connected(item)
        segments_set.append(item)
      elif isinstance(item, gfapy.line.edge.GFA2):
        self._check_induced_set_elem_connected(item)
        for sl in [item.sid1.line, item.sid2.line]:
          self._check_induced_set_elem_connected(sl)
          segments_set.append(sl)
      elif isinstance(item, gfapy.line.group.Ordered):
        self._check_induced_set_elem_connected(item)
        subset = item.captured_segments
        assert(subset)
        for elem in subset:
          segments_set.append(elem.line)
      elif isinstance(item, gfapy.line.group.Unordered):
        self._check_induced_set_elem_connected(item)
        subset = item.induced_segments_set
        assert(subset)
        for elem in subset:
          segments_set.append(elem)
      elif isinstance(item, gfapy.line.Unknown):
        raise gfapy.RuntimeError(
          "Induced set cannot be computed; a reference has not been resolved\n"+
          "Line: {}\n".format(self)+
          "Unresolved reference: {} (Virtual unknown line)".format(item.name))
      else:
        raise gfapy.TypeError(
          "Line: {}\t".format(self)+
          "Cannot compute induced set:\t"+
          "Error: items of type {} are not supported\t".format(item.__class__.__name__)+
          "Unsupported item: {}".format(item))
    unique_ids = set()
    return [e for e in segments_set \
        if id(e) not in unique_ids and not unique_ids.add(id(e))]

  def _check_induced_set_elem_connected(self, item):
    if not item.is_connected():
      raise gfapy.RuntimeError(
        "Cannot compute induced set\n"+
        "Non-connected element found\n"+
        "Item: {}\nLine: {}".format(item, self))

  def _compute_induced_edges_set(self, segments_set):
    edges_set = list()
    for item in segments_set:
      for edge in item.edges:
        if edge.other(item) in segments_set:
          edges_set.append(edge)
    unique_ids = set()
    return [e for e in edges_set \
        if id(e) not in unique_ids and not unique_ids.add(id(e))]



import gfapy

class FieldData:

  @property
  def positional_fieldnames(self):
    """Name of the positional fields.
    Returns:
      str list

    Note:
      The field names can vary from the exact name of the
      fields in the specification, due to compatibility issues.
    """
    return self.__class__.POSFIELDS

  @property
  def tagnames(self):
    """Names of the tags defined in the line.
    Returns:
      str list
    """
    return [ x for x in self._data.keys() \
        if x not in self.positional_fieldnames ]

  @property
  def record_type(self):
    """The record type.

    The content of the first field of the line. In all predefined record types
    this is a single-letter upper case string.

    Returns:
      str
    """
    return self.__class__.RECORD_TYPE

  def set(self, fieldname, value):
    """Set the value of a field.

    If a datatype for a new custom tag is not set,
    the default for the value assigned to the field will be used
    (e.g. J for Hashes, i for Integer, etc).

    Parameters
    ----------
    fieldname : str
      The name of the field to set.
      (positional field, predefined or custom tag)

    Raises
    ------
    gfapy.FormatError
      If **fieldname** is not a valid predefined or
      custom tag name (and **validate["tags"]**).

    Returns
    -------
    object
      **value**
    """
    if fieldname in self._data or self._is_predefined_tag(fieldname):
      return self._set_existing_field(fieldname, value)
    elif fieldname in self.__class__.FIELD_ALIAS:
      return self.set(self.__class__.FIELD_ALIAS[fieldname], value)
    elif self.virtual:
      raise gfapy.RuntimeError("Virtual lines do not have tags")
    elif (self.vlevel == 0) or self._is_valid_custom_tagname(fieldname):
      self._define_field_methods(fieldname)
      if self._datatype.get(fieldname, None) is not None:
        return self._set_existing_field(fieldname, value)
      elif value is not None:
        self._datatype[fieldname] = \
            gfapy.Field._get_default_gfa_tag_datatype(value)
        self._data[fieldname] = value
        return self._data[fieldname]
    else:
      raise gfapy.FormatError(
        "{} is not a positional field,".format(fieldname)+
        "an existing tag, an alias, a predefined tag or a valid custom tag\n"+
        "positional fields: {}\n".format(", ".join(self.positional_fieldnames))+
        "existing tags: {}\n".format(", ".join(self.tagnames))+
        "aliases: {}\n".format(", ".join(self.__class__.FIELD_ALIAS.keys()))+
        "predefined tags: {}\n".format(", ".join(self.__class__.PREDEFINED_TAGS)))

  def get(self, fieldname):
    """
    Get the value of a field.

    Parameters
    ----------
    fieldname : str
      Name of the field.

    Returns
    -------
    object or None
      Value of the field or **None** if field is not defined.
    """
    v = self._data.get(fieldname, None)
    if isinstance(v, str):
      t = self._field_datatype(fieldname)
      if t != "Z" and t != "seq":
        # value was not parsed or was set to a string by the user
        self._data[fieldname] = gfapy.Field._parse_gfa_field(v, t,
                                                    safe = (self.vlevel >= 1),
                                                    fieldname = fieldname,
                                                    line = self)
        return self._data[fieldname]
      else:
        if (self.vlevel >= 3):
          gfapy.Field._validate_gfa_field(v, t, fieldname)
    elif v is not None:
      if (self.vlevel >= 3):
        t = self._field_datatype(fieldname)
        gfapy.Field._validate_gfa_field(v, t, fieldname)
    else:
      dealiased_fieldname = self.__class__.FIELD_ALIAS.get(fieldname, None)
      if dealiased_fieldname is not None:
        return self.get(dealiased_fieldname)
    return v

  def try_get(self, fieldname):
    """
    Value of a field, raising an exception if it is not defined.

    Parameters
    ----------
    fieldname : str
      Name of the field.

    Raises
    ------
    gfapy.NotFoundError
      If field is not defined.

    Returns
    -------
    object or None
      Value of the field.
    """
    v = self.get(fieldname)
    if v is None:
      raise gfapy.NotFoundError(
        "No value defined for tag {}".format(fieldname))
    return v

  def delete(self, tagname):
    """
    Remove a tag from the line, if it exists; do nothing if it does not.

    Parameters
    ----------
    tagname : Symbol
      The tag name of the tag to remove.

    Returns
    -------
    object or None
      The deleted value or None, if the field was not defined.
    """
    if tagname in self.tagnames:
      if tagname in self._datatype:
        self._datatype.pop(tagname)
      return self._data.pop(tagname)
    else:
      return None

  def _set_existing_field(self, fieldname, value, set_reference = False):
    renaming_connected = False
    if self._gfa:
      if not set_reference and \
        (fieldname in self.__class__.REFERENCE_FIELDS or \
         fieldname in self.__class__.BACKREFERENCE_RELATED_FIELDS):
        raise gfapy.RuntimeError(
          "The value of field '{}' cannot be changed, ".format(fieldname)+
          "as the line belongs to a GFA instance")
      if (fieldname == self.__class__.STORAGE_KEY) or \
        (self.__class__.STORAGE_KEY == "name" and \
        fieldname == self.__class__.NAME_FIELD):
         renaming_connected = True
         self._gfa._unregister_line(self)
    if value is None:
      if fieldname in self._data:
        self._data.pop(fieldname)
    else:
      if self.vlevel >= 3:
        self._field_or_default_datatype(fieldname, value)
        gfapy.Field._validate_gfa_field(value, self._field_datatype(fieldname),
            fieldname)
      self._data[fieldname] = value
    if renaming_connected:
      self._gfa._register_line(self)

  def _dealias_fieldname(self, fieldname):
    return self.__class__.FIELD_ALIAS.get(fieldname, fieldname)

  def _dealias_fieldnames(self, fieldnames):
    fieldnames[:] = map(self._dealias_fieldname, fieldnames)



import gfapy
import unittest

class TestGraphopCopyNumber(unittest.TestCase):

  def test_delete_low_coverage_segments(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/copynum.1.{}".format(sfx))
      self.assertEqual({"0","1","2"}, set(gfa.segment_names))
      gfa.delete_low_coverage_segments(10)
      self.assertEqual({"1","2"}, set(gfa.segment_names))
      gfa.delete_low_coverage_segments(100)
      self.assertEqual({"2"}, set(gfa.segment_names))
      gfa.delete_low_coverage_segments(1000)
      self.assertEqual(set(), set(gfa.segment_names))

  def test_compute_copy_numbers(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/copynum.2.{}".format(sfx))
      gfa.compute_copy_numbers(9) # nothing raised
      self.assertEqual(0, gfa.try_get_segment("0").cn)
      self.assertEqual(1, gfa.try_get_segment("1").cn)
      self.assertEqual(2, gfa.try_get_segment("2").cn)
      self.assertEqual(3, gfa.try_get_segment("3").cn)

  def test_apply_copy_number(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/copynum.2.{}".format(sfx))
      self.assertEqual({"0","1","2","3"}, set(gfa.segment_names))
      gfa.compute_copy_numbers(9)
      gfa.apply_copy_numbers()
      self.assertEqual({"1","2","3","2*2","3*2","3*3"}, set(gfa.segment_names))
      gfa.compute_copy_numbers(9)
      assert(all(x.cn == 1 for x in gfa.segments))




import unittest
import gfapy

class TestUnitAlignment(unittest.TestCase):

  cigar_1 = gfapy.CIGAR([
      gfapy.CIGAR.Operation(12,"M"),
      gfapy.CIGAR.Operation(1,"D"),
      gfapy.CIGAR.Operation(2,"I"),
      gfapy.CIGAR.Operation(0,"M"),
      gfapy.CIGAR.Operation(1,"P")])

  cigar_1_a = [
     gfapy.CIGAR.Operation(12,"M"),
     gfapy.CIGAR.Operation(1,"D"),
     gfapy.CIGAR.Operation(2,"I"),
     gfapy.CIGAR.Operation(0,"M"),
     gfapy.CIGAR.Operation(1,"P")]

  cigar_1_s = "12M1D2I0M1P"

  trace_1 = gfapy.Trace([12,12,0])
  trace_1_s = "12,12,0"
  trace_1_a = [12,12,0]

  def test_list_to_alignment(self):
    assert(isinstance(gfapy.Alignment([]),gfapy.AlignmentPlaceholder))
    self.assertEqual(TestUnitAlignment.cigar_1, gfapy.Alignment(TestUnitAlignment.cigar_1_a))
    self.assertRaises(gfapy.VersionError, gfapy.Alignment, TestUnitAlignment.trace_1_a, version="gfa1")
    self.assertEqual(TestUnitAlignment.trace_1, gfapy.Alignment(TestUnitAlignment.trace_1_a, version="gfa2"))
    self.assertRaises(gfapy.VersionError, gfapy.Alignment, TestUnitAlignment.cigar_1_a, version="gfaX")
    self.assertRaises(gfapy.FormatError, gfapy.Alignment, ["x",2,1])
    # only the first element is checked, therefore:
    malformed1 = [1,2,"x"]
    gfapy.Alignment(malformed1, version="gfa2") # nothing raised
    assert(isinstance(gfapy.Alignment(malformed1, version="gfa2"), gfapy.Trace))
    self.assertRaises(gfapy.TypeError, gfapy.Alignment(malformed1, version="gfa2").validate)
    malformed2 = [gfapy.CIGAR.Operation(12,"M"),2,"x"]
    gfapy.Alignment(malformed2) # nothing raised
    assert(isinstance(gfapy.Alignment(malformed2),gfapy.CIGAR))
    self.assertRaises(gfapy.TypeError, gfapy.Alignment(malformed2).validate)

  def test_cigar_from_string(self):
    self.assertEqual(TestUnitAlignment.cigar_1,
        gfapy.CIGAR._from_string(TestUnitAlignment.cigar_1_s))
    assert(isinstance(gfapy.CIGAR._from_string("*"),
      gfapy.AlignmentPlaceholder))
    self.assertEqual(TestUnitAlignment.cigar_1,
        gfapy.CIGAR(TestUnitAlignment.cigar_1_a))

  def test_trace_from_string(self):
    self.assertEqual(TestUnitAlignment.trace_1,
        gfapy.Trace._from_string(TestUnitAlignment.trace_1_s))
    self.assertRaises(gfapy.FormatError, gfapy.Trace._from_string, "A,1,2")




import gfapy

class NumericArray(list):
  """
  A numeric array representable using the data type B of the GFA specification.
  """

  SIGNED_INT_SUBTYPE = ["c", "s", "i"]
  """
  Subtypes for signed integers, from the smallest to the largest
  """

  UNSIGNED_INT_SUBTYPE = [ st.upper() for st in SIGNED_INT_SUBTYPE ]
  """
  Subtypes for unsigned integers, from the smallest to the largest
  """

  INT_SUBTYPE = UNSIGNED_INT_SUBTYPE + SIGNED_INT_SUBTYPE
  """
  Subtypes for integers
  """

  FLOAT_SUBTYPE = ["f"]
  """
  Subtypes for floats
  """

  SUBTYPE = INT_SUBTYPE + FLOAT_SUBTYPE
  """
  Subtypes
  """

  SUBTYPE_BITS = {"c" : 8, "s" : 16, "i" : 32}
  """
  Number of bits of unsigned integer subtypes
  """

  SUBTYPE_RANGE = {
    "C" : (0, 2**8),
    "S" : (0, 2**16),
    "I" : (0, 2**32),
    "c" : (-(2**(8 - 1)), 2**(8 - 1)),
    "s" : (-(2**(16 - 1)), 2**(16 - 1)),
    "i" : (-(2**(32 - 1)), 2**(32 - 1))
  }
  """
  Range for integer subtypes
  (Python-style, i.e. range[1] is not included)
  """

  def validate(self):
    """
    Validate the numeric array

    Raises
    ------
    gfapy.ValueError
      If the array is not valid
    """
    self.compute_subtype()

  def compute_subtype(self):
    """
    Computes the subtype of the array from its content.

    If all elements are float, then the computed subtype is "f".
    If all elements are integer, the smallest possible numeric subtype
    is computed; thereby,
    if all elements are non-negative, an unsigned subtype is selected,
    otherwise a signed subtype.
    In all other cases an exception is raised.

    Raises
    ------
    gfapy.ValueError
      If the array is not a valid numeric array

    Returns
    -------
    one of gfapy.NumericArray.SUBTYPE
    """
    if all([ isinstance(f, float) for f in self]):
      return "f"
    else:
      e_max = None
      e_min = None
      for e in self:
        if not isinstance(e, int):
          raise gfapy.ValueError(
            "NumericArray does not contain homogenous numeric values\n"+
            "Content: {}".format(repr(self)))
        if (e_max is None or e > e_max): e_max = e
        if (e_min is None or e < e_min): e_min = e
      return gfapy.NumericArray.integer_type((e_min,e_max))

  @staticmethod
  def integer_type(range):
    """
    Computes the subtype for integers in a given range.

    If all elements are non-negative, an unsigned subtype is selected,
    otherwise a signed subtype.

    Parameters
    ----------
    range : (int, int)
      The integer range (min, max)

    Raises
    ------
    gfapy.ValueError
      If the integer range is outside all subtype ranges

    Returns
    -------
    one of gfapy.NumericArray.INT_SUBTYPE
      subtype code
    """
    if range[0] < 0:
      for st in NumericArray.SIGNED_INT_SUBTYPE:
        st_range = NumericArray.SUBTYPE_RANGE[st]
        if st_range[0] <= range[0] and st_range[1] > range[1]:
          return st
    else:
      for st in NumericArray.UNSIGNED_INT_SUBTYPE:
        st_range = NumericArray.SUBTYPE_RANGE[st]
        if st_range[1] > range[1]:
          return st
    raise gfapy.ValueError(
      "NumericArray: values are outside of all integer subtype ranges\n"+
      "Range: {}".format(repr(range)))

  def __str__(self):
    """
    GFA datatype B representation of the numeric array

    Raises
    ------
    gfapy.ValueError
      if the array is not a valid numeric array

    Returns
    -------
    str
    """
    subtype = self.compute_subtype()
    return "{},{}".format(subtype, ",".join([str(v) for v in self]))

  def _default_gfa_tag_datatype(self):
    """
    GFA tag datatype to use, if none is provided

    Returns
    -------
    one of gfapy.Field.TAG_DATATYPE
    """
    return "B"

  @classmethod
  def from_string(cls, string, valid = False):
    """
    Create a numeric array from a string

    Parameters
    ----------
    string : str
    valid : optional bool
      *(default:* **False** *)*
      If **False**, validate the range of the numeric values, according
      to the array subtype. If **True** the string is guaranteed to be valid.

    Raises
    ------
    gfapy.ValueError
      If any value is not compatible with the subtype.
    gfapy.TypeError
      If the subtype code is invalid.

    Returns
    -------
    gfapy.NumericArray
      The numeric array
    """
    if not valid:
      if len(string) == 0:
        raise gfapy.FormatError("Numeric array string shall not be empty")
      if string[-1] == ",":
        raise gfapy.FormatError("Numeric array string ends with comma\n"+
          "String: {}".format(string))
    elems = string.split(",")
    subtype = elems[0]
    if subtype not in NumericArray.SUBTYPE:
      raise gfapy.TypeError("Subtype {} unknown".format(subtype))
    if subtype != "f":
      range = NumericArray.SUBTYPE_RANGE[subtype]
    def gen():
      for e in elems[1:]:
        if subtype != "f":
          try:
            e = int(e)
          except:
            raise gfapy.ValueError("Value is not valid: {}\n".format(e)+
                "Numeric array string: {}".format(string))
          if not valid and not (e >= range[0] and e < range[1]):
            raise gfapy.ValueError((
                    "NumericArray: "+
                    "value is outside of subtype {0} range\n"+
                    "Value: {1}\n"+
                    "Range: {2}\n"+
                    "Content: {3}").format(subtype, e,
                      repr(range), repr(elems)))
          yield e
        else:
          yield float(e)
    return cls(list(gen()))



import gfapy

class Disconnection:

  def disconnect(self):
    """
    Remove the line from the GFA instance it belongs to, if any.

    The Line instance itself will still exist, but all references from it to
    other lines are deleted, as well as references to it from other lines.
    Mandatory references are turned into their non-reference representations
    (e.g. segments references in the sid fields of E lines
    or in the from/to lines of L/C lines are changed into symbols).
    """
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Line {} is not connected to a GFA instance".format(self))
    self._remove_field_backreferences()
    self._remove_field_references()
    self._disconnect_dependent_lines()
    self._remove_nonfield_backreferences()
    self._remove_nonfield_references()
    self._gfa._unregister_line(self)
    self._gfa = None

  def _delete_reference(self, line, key):
    if key not in self._refs: return
    idx = None
    for i, x in enumerate(self._refs[key]):
      if x is line:
        idx = i
    if idx is None:
      return
    elif idx == 0:
      self._refs[key].pop(0)
    elif idx == len(self._refs[key])-1:
      self._refs[key].pop()
    else:
      self._refs[key] = (self._refs[key][0:idx] + self._refs[key][idx+1:])

  def _delete_first_reference(self, key):
    if not self._refs or not self._refs[key]:
      return
    self._refs[key].pop(0)

  def _delete_last_reference(self, key):
    if not self._refs or not self._refs[key]:
      return
    self._refs[key].pop()

  def _remove_field_references(self):
    """
    .. note::
      currently this method supports fields which are: references,
      oriented lines and lists of references of oriented lines;
      if SUBCLASSES have reference fields which contain references
      in a different fashion, the method must be updated or overwritten
      in the subclass
    """
    for k in self.__class__.REFERENCE_FIELDS:
      ref = self.get(k)
      if isinstance(ref, gfapy.Line):
        self._set_existing_field(k, ref.name, set_reference = True)
      elif isinstance(ref, gfapy.OrientedLine):
        ref.line = ref.name
      elif isinstance(ref, list):
        for i, elem in enumerate(ref):
          if isinstance(elem, gfapy.Line):
            ref[i] = elem.name
          elif isinstance(elem, gfapy.OrientedLine):
            ref[i].line = elem.name

  def _remove_backreference(self, ref, k):
    if isinstance(ref, gfapy.Line):
      ref._update_references(self, None, k)
    elif isinstance(ref, gfapy.OrientedLine):
      if isinstance(ref.line, gfapy.Line):
        ref.line._update_references(self, None, k)
    elif isinstance(ref, list):
      for i in range(len(ref)):
       self._remove_backreference(ref[i], k)

  def _disconnect_dependent_line(self, ref):
    if isinstance(ref, gfapy.Line):
      ref.disconnect()
    elif isinstance(ref, gfapy.OrientedLine):
      if isinstance(ref.line, gfapy.Line):
        ref.line.disconnect()
    elif isinstance(ref, list):
      for i in range(len(ref)):
        self._disconnect_dependent_line(ref[i])

  def _remove_field_backreferences(self):
    """
    .. note::
      currently this method supports fields which are: references,
      oriented lines and lists of references of oriented lines;
      if SUBCLASSES have reference fields which contain references
      in a different fashion, the method must be updated or overwritten
      in the subclass
    """
    for k in self.__class__.REFERENCE_FIELDS:
      self._remove_backreference(self.get(k), k)

  def _disconnect_dependent_lines(self):
    for k in self.__class__.DEPENDENT_LINES:
      for ref in self._refs.get(k, []):
        self._disconnect_dependent_line(ref)

  def _remove_nonfield_backreferences(self):
    for k in self.__class__.OTHER_REFERENCES:
      for ref in self._refs.get(k, []):
        self._remove_backreference(ref, k)

  def _remove_nonfield_references(self):
    self._refs = {}



import gfapy
import builtins
from .validator import Validator
from .parser import Parser
from .writer import Writer
from . import alignment_gfa1                as field_alignment_gfa1
from . import alignment_gfa2                as field_alignment_gfa2
from . import alignment_list_gfa1           as field_alignment_list_gfa1
from . import byte_array                    as field_byte_array
from . import char                          as field_char
from . import comment                       as field_comment
from . import custom_record_type            as field_custom_record_type
from . import float                         as field_float
from . import generic                       as field_generic
from . import identifier_gfa2               as field_identifier_gfa2
from . import oriented_identifier_gfa2      as field_oriented_identifier_gfa2
from . import identifier_list_gfa2          as field_identifier_list_gfa2
from . import integer                       as field_integer
from . import json                          as field_json
from . import numeric_array                 as field_numeric_array
from . import optional_identifier_gfa2      as field_optional_identifier_gfa2
from . import optional_integer              as field_optional_integer
from . import orientation                   as field_orientation
from . import oriented_identifier_list_gfa1 as field_oriented_identifier_list_gfa1
from . import oriented_identifier_list_gfa2 as field_oriented_identifier_list_gfa2
from . import path_name_gfa1                as field_path_name_gfa1
from . import position_gfa1                 as field_position_gfa1
from . import position_gfa2                 as field_position_gfa2
from . import segment_name_gfa1             as field_segment_name_gfa1
from . import sequence_gfa1                 as field_sequence_gfa1
from . import sequence_gfa2                 as field_sequence_gfa2
from . import string                        as field_string

class Field(Validator, Parser, Writer):
  """
  Support for the decoding, validation and encoding of data in GFA fields.

  Classes are defined (and imported here) for each type of field (positional and
  tags) defined in the GFA specifications. The field definition classes
  implement the following methods:

  * ``decode(str)``: decodes the content of a GFA field into a Python object
      which represents the value; the content of the field is validated and
      the returned object is guaranteed to be valid
  * ``unsafe_decode(str)``: an optional method, which decodes the content
      of the string faster than decode(), but does not perform validations
  * ``validate_encoded(str)``: validates the content of a GFA field, when this
      is in its string form; it can be called by the decode() method
  * ``validate_decoded(obj)``: validates the content of a GFA field, when this
      is in a non-string form; it can be called by the encode() method
  * ``encode(obj)``: takes a non-string content of a GFA field and converts it
      in its string representation according to the GFA specification; the
      returned string is guaranteed to be valid
  * ``unsafe_encode(obj)``: an optional method, which encodes the content
      of a non-string field faster than encode(), but does not perform
      validations

  Notes:
    The library user does not call these methods directly, as the interaction
    is done using the interface of the `~gfapy.line.line.Line` class.
    However, an user may define classes for custom datatypes, to be used with
    custom record types.
  """

  _default_tag_datatypes = [
    (builtins.int , "i"),
    (builtins.float , "f"),
    (builtins.dict , "J"),
    (builtins.list , "J"),
    (builtins.object , "Z")
  ]
  """Default tag datatype to be used if the value is of a built-in class.

  For non build-in classes, the _default_gfa_tag_datatype() method of the
  class is called instead.
  """

  GFA1_POSFIELD_DATATYPE = [
                             "alignment_gfa1",
                             "alignment_list_gfa1",
                             "oriented_identifier_list_gfa1",
                             "position_gfa1",
                             "segment_name_gfa1",
                             "sequence_gfa1",
                             "path_name_gfa1",
                           ]
  """The names of the GFA1-specific datatypes for positional fields."""

  GFA2_POSFIELD_DATATYPE = [
                             "alignment_gfa2",
                             "generic",
                             "identifier_gfa2",
                             "oriented_identifier_gfa2",
                             "identifier_list_gfa2",
                             "oriented_identifier_list_gfa2",
                             "optional_identifier_gfa2",
                             "position_gfa2",
                             "custom_record_type",
                             "sequence_gfa2",
                             "optional_integer",
                           ]
  """The names of the GFA2-specific datatypes for positional fields."""

  GFAX_POSFIELD_DATATYPE = [ "comment", "orientation" ]
  """The names of the non version-specific datatypes for positional fields."""

  POSFIELD_DATATYPE = GFA1_POSFIELD_DATATYPE + \
                      GFA2_POSFIELD_DATATYPE + \
                      GFAX_POSFIELD_DATATYPE
  """The names of all datatypes for positional fields."""

  TAG_DATATYPE = ["A", "i", "f", "Z", "J", "H", "B"]
  """The names of all datatypes for tags."""

  FIELD_DATATYPE = TAG_DATATYPE + POSFIELD_DATATYPE
  """The names of all datatypes for positional fields and tags."""

  FIELD_MODULE = {
    "alignment_gfa1"                 : field_alignment_gfa1,
    "alignment_gfa2"                 : field_alignment_gfa2,
    "alignment_list_gfa1"            : field_alignment_list_gfa1,
    "byte_array"                     : field_byte_array,
    "char"                           : field_char,
    "comment"                        : field_comment,
    "custom_record_type"             : field_custom_record_type,
    "float"                          : field_float,
    "generic"                        : field_generic,
    "identifier_gfa2"                : field_identifier_gfa2,
    "oriented_identifier_gfa2"       : field_oriented_identifier_gfa2,
    "identifier_list_gfa2"           : field_identifier_list_gfa2,
    "integer"                        : field_integer,
    "json"                           : field_json,
    "numeric_array"                  : field_numeric_array,
    "optional_identifier_gfa2"       : field_optional_identifier_gfa2,
    "optional_integer"               : field_optional_integer,
    "orientation"                    : field_orientation,
    "oriented_identifier_list_gfa1"  : field_oriented_identifier_list_gfa1,
    "oriented_identifier_list_gfa2"  : field_oriented_identifier_list_gfa2,
    "path_name_gfa1"                 : field_path_name_gfa1,
    "position_gfa1"                  : field_position_gfa1,
    "position_gfa2"                  : field_position_gfa2,
    "segment_name_gfa1"              : field_segment_name_gfa1,
    "sequence_gfa1"                  : field_sequence_gfa1,
    "sequence_gfa2"                  : field_sequence_gfa2,
    "string"                         : field_string,
    "H"    : field_byte_array,
    "A"    : field_char,
    "f"    : field_float,
    "i"    : field_integer,
    "J"    : field_json,
    "B"    : field_numeric_array,
    "Z"    : field_string,
  }
  """Assignment of a class for the parsing, validation and encoding of data.

  The dictionary contains keys for each GFA datatype; the value is a class name,
  which provides the encoding, decoding and validation methods.

  For simplicity of use, tag datatypes are present twice, once with a
  one-letter symbol (such as i) and once with a longer labe; (such as integer).
  """

  # Returns the default GFA tag for the given object.
  @staticmethod
  def _get_default_gfa_tag_datatype(obj):
    """Default GFA tag datatype for a given object

    Parameters:
      obj : an object of any Python class

    Returns:
      str : the identifier of a datatype (one of the keys of FIELD_MODULE)
        to be used for a tag with obj as value, if a datatype has not
        been specified by the user
    """
    if getattr(obj, "_default_gfa_tag_datatype",None):
      return obj._default_gfa_tag_datatype()
    else:
      if isinstance(obj, list) and\
             (all([isinstance(v, builtins.int) for v in obj]) or
              all([isinstance(v, builtins.float) for v in obj])):
        return "B"
      for k,v in gfapy.Field._default_tag_datatypes:
        if isinstance(obj, k):
          return v
      return "J"

  @classmethod
  def register_datatype(cls, name, klass):
    """Register a custom-defined datatype class

    Parameters:
      name (str) : the identifier to be used for the datatype. This is
        to be used in the field datatype declaration of extensions
        definining custom records, which use this custom datatype
      klass (class) : the class which provide the decode, encode,
        validate_encoded and validate_decoded methods for handling
        data of the custom datatype
    """
    cls.GFA2_POSFIELD_DATATYPE.append(name)
    cls.FIELD_MODULE[name] = klass




import gfapy
import unittest

class TestApiGFA2Lines(unittest.TestCase):

  def test_S(self):
    fields=["S","1","ACGTCACANNN","RC:i:1232","LN:i:11","ab:Z:abcd",
            "FC:i:2321","KC:i:1212"]
    s="\t".join(fields)
    gfapy.Line(s) # nothing raised
    self.assertEqual(gfapy.line.segment.GFA1, gfapy.Line(s).__class__)
    self.assertEqual(fields[0], gfapy.Line(s).record_type)
    self.assertEqual(fields[1], gfapy.Line(s).name)
    self.assertEqual(fields[2], gfapy.Line(s).sequence)
    self.assertEqual(1232, gfapy.Line(s).RC)
    self.assertEqual(11, gfapy.Line(s).LN)
    self.assertEqual(2321, gfapy.Line(s).FC)
    self.assertEqual(1212, gfapy.Line(s).KC)
    self.assertEqual("abcd", gfapy.Line(s).ab)
    with self.assertRaises(gfapy.FormatError): s+gfapy.Line("\tH1")
    with self.assertRaises(gfapy.FormatError): gfapy.Line("S\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[2]="!@#?"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.TypeError):
      f=fields.copy(); f[3]="RC:Z:1232"; gfapy.Line("\t".join(f),version="gfa1")
    f=["S","2","ACGTCACANNN","LN:i:3"]
    with self.assertRaises(gfapy.InconsistencyError):
      gfapy.Line("\t".join(f),vlevel=1, version="gfa1")
    f=["S","2","ACGTCACANNN","LN:i:11"]
    gfapy.Line("\t".join(f)) # nothing raised
    f=["S","2","*","LN:i:3"]
    gfapy.Line("\t".join(f)) # nothing raised

  def test_coverage(self):
    l = gfapy.Line("S\t0\t*\tRC:i:600\tLN:i:100")
    self.assertEqual(6, l.coverage())
    self.assertEqual(6, l.try_get_coverage())
    l = gfapy.Line("S\t0\t*\tRC:i:600")
    self.assertEqual(None, l.coverage())
    with self.assertRaises(gfapy.NotFoundError): l.try_get_coverage()
    l = gfapy.Line("S\t0\t*\tLN:i:100")
    self.assertEqual(None, l.coverage())
    with self.assertRaises(gfapy.NotFoundError): l.try_get_coverage()
    l = gfapy.Line("S\t0\t*\tFC:i:600\tLN:i:100")
    self.assertEqual(None, l.coverage())
    with self.assertRaises(gfapy.NotFoundError): l.try_get_coverage()
    self.assertEqual(6, l.coverage(count_tag="FC"))
    self.assertEqual(6, l.try_get_coverage(count_tag="FC"))

